/*****************************************************************/
/*                                                               */
/*   dmmask.c                                                    */
/*                                                               */
/*   Functions that work with PixelMask                          */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmlibP.h"
#include "../../region/src/region_priv.h" 

#define DM_MASK_NMAX_DEFAULT 20

/**************************************/
/* function called in DM modules      */
/**************************************/

regRegion* dmMaskParse( char* regstring);
dmErrCode  dmMaskWrite( dmDataset* ds, char* name, dmMask* newmask);
char*      dmMaskName( void );
int        dmi_mask_remove_blkname(char** ss_string);

regRegion* dmi_mask_region_from_string( dmDescriptor* ssd, const char* dsref, const char* dsval );
void       dmi_mask_get_transforms(     dmDescriptor* desc, dmCoordTransform** log2phys, dmCoordTransform** phys2wcs );
dmBool     dmi_mask_match_coords( regShape* shape, dmBlock *src_blk, char* cptname);
regRegion* dmi_mask_rectangle( char* regstring );

/**************************************/
/* bridge functions to region-lib     */
/**************************************/

int       dmi_mask_inside(  regShape *shape, double xpos, double ypos);
int       dmi_mask_extent(  regShape *shape, double *xpos, double *ypos );
regShape* dmi_mask_copy(    regShape *inshp);
double    dmi_mask_area(    regShape *shape );
int       dmi_mask_isequal( regShape *shape1, regShape *shape2);
void      dmi_mask_tostring(regShape *shape, char* buf, long maxlen );
void      dmi_mask_free(    regShape *shape);

/**************************************/
/* MASKs Consolidation functions      */
/**************************************/
struct dmMask_s 
{
  char *blkname;               /* name string of mask DM block           */

  long   ndim;                 /* Dimension of data array                */
  long*  axlens;               /* Lengths of data array axes.            */
  unsigned char* data;         /* Mask data array in bytes, 1||0         */
  int cflag;                   /* mask block coordinate flag, PHYS|LOGIC, dynamically updated possible */

  dmCoordTransform* log2phys;  /* Transform from Logical->Physical       */
  dmCoordTransform* phys2wcs;  /* Transform from Physical->World         */
  char **l_cptNames;           /* cptNames of Linear coord transformation, say X and Y of "sky(x,y)" */
  char **w_cptNames;           /* cptNames of non-Linear coord transformation, say RA and DEC of "EQPO(RA, DEC)" */

  double* xbounds;             /* Mask bounds in mask cflag indicated coords - Xaxis   */
  double* ybounds;             /* Mask bounds in mask cflag indicated coords - Yaxis   */
};
// typedef struct dmMask_s dmMask;  defined in ../misc/dmlibP.h

/**************************************/
/* all local (Mask) functions      */
/**************************************/

/* interface via dmBlock */
char*       dmi_mask_check_syntax(   char* ptr, int* incl );
regRegion*  dmi_mask_read_mask(      dmBlock* msk_blk );
dmBool      dmi_mask_isvalid_mask(   dmBlock* msk_blk );
void*       dmi_mask_get_image_array(dmBlock* msk_blk,  dmDataType* dtype, long* imgsz);
dmMask*     dmi_mask_convert(        dmBlock *msk_blk);

void*       dmi_mask_get_null(       dmDescriptor *dd);
double*     dmi_mask_get_image_field(dmDescriptor* dd);
dmErrCode   dmi_mask_copy_coordsys(  dmDescriptor* dd, dmMask* newmask);


/* interface via Mask-struct */
dmBool      dmi_mask_isvalid_transf( regShape* shape, dmDescriptor* src_dd);
void        dmi_mask_set_includeflag(regShape* shape, int fincl);
void        dmi_mask_set_coordflag(regRegion* region, int fcoord);

dmMask*    dmi_mask_consolidate( regRegion* region, double* field );
regShape*  dmi_mask_find_masks(  regRegion* region, long* nmasks );
void*      dmi_mask_convert_values(void* values, dmDataType dtype, long nvals, void* nullValue  );
regRegion* dmi_mask_list_to_region( regShape* inmasks );

void*      dmi_mask_get_spec(regShape* shape);
void       dmi_mask_set_spec(regShape* shape, void* smask);
dmMask*    dmi_mask_maskstruct_alloc(int ndim) ;
dmMask*    dmi_mask_maskstruct_copy(dmMask* src);
dmBool     dmi_mask_maskstruct_isequal(dmMask *smask1, dmMask* smask2);
int        dmi_mask_maskstruct_invert_bounds(dmMask *smask, int invert);
void       dmi_mask_maskstruct_free(  dmMask* smask);
void       dmi_mask_maskstruct_print( dmMask *smask );

/* Rebinning */
dmBool dmi_mask_match_bin(  dmSubspace* subspace, dmBool isbinned);
dmBool dmi_mask_match_scale_multiple(regShape *myShape, dmDescriptor *src_dd, long *multis);
int    dmi_mask_rebin_mask( dmSubspace* subspace, dmBinSpec* binspec);
int    dmi_mask_rebin_data( regShape *shape,      dmBinSpec*  binspec);

/* Utilities */
void  dmi_mask_print_transform( dmCoordTransform* t );
void  dmi_mask_print_image(long xlen, long ylen, unsigned char* data);

/**********************************************************************/
/*                                                                    */
/* Function:  dmi_mask_set_coordflag                                  */
/*                                                                    */
/* Description:                                                       */
/*   Update mask coordinate flag.                                     */
/*                                                                    */
/*  Input parameters:                                                 */
/*    region: region contains mask shape which is defined by coodflag */
/*     cflag: coord flag to be assigned to the mask's                 */
/*                                                                    */
/*  Output parameters:                                                */
/*    region's mask's cood-flag is updated                            */
/*                                                                    */
/*  Returns:                                                          */
/*    none                                                            */
/*                                                                    */
/**********************************************************************/
void dmi_mask_set_coordflag(regRegion* region, int cflag)
{   
  regShape* iptr = NULL;   
  iptr = region->shape; 

  while ( iptr != NULL ) { 
    if ( iptr->type == regMASK ) {
      iptr->flag_coord = cflag;
    }
    iptr = iptr->next;
  }
}


/********************************************************************/
/*                                                                  */
/* Function:  dmMaskName                                            */
/*                                                                  */
/* Description:                                                     */
/*   Provides the 'name' of Mask region shape.                      */
/*    for use in region string                                      */
/*        MASK(blkname)                                             */
/*      or subspace block name                                      */
/* Note:                                                            */
/*   User must free the returned memory.                            */
/*                                                                  */
/*  Input parameters:                                               */
/*    none                                                          */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
/*  Returns:                                                        */
/*    char* name      - name of mask region shape                   */
/*                                                                  */
/********************************************************************/
char* dmMaskName(void)
{
  char *buff=calloc(5, sizeof(char) );
  strcpy(buff, "MASK");
  return buff;
}
/***************************************************************/
/*                                                             */
/* Function:   dmi_mask_rectangle                              */
/*                                                             */
/* Description:                                                */
/*   Support "bounds(mask())"  syntax.                         */
/*   get the mask-file's tlmin/tlmax as the bounds of          */
/*   a rectangle and return the shape of "RECTANGLE" to        */
/*   overwrite the "MASK".                                     */
/*                                                             */
/*  Input parameters:                                          */
/*    char             *regstring                              */
/*                                                             */
/*  Output parameters:                                         */
/*    none                                                     */
/*                                                             */
/*  Returns:                                                   */
/*          regRegion  *region                                 */
/*                                                             */
/***************************************************************/
regRegion* dmi_mask_rectangle( char* regstring)
{
  regRegion* region;
  char *string = regstring;
  int extent = 0;
  int is_reg = 0;
  double* fieldx=NULL;
  double* fieldy=NULL;
 
  /* Support "[sky=bounds(mask(my.mask))]" syntax */
  string = dmi_region_check_bounds_syntax( regstring, &extent, &is_reg );
  if( string ){
    /* create 'Mask' struct and store the bounds */
    region=dmMaskParse(string);
    free(string);
  } 
  
  if ( region && extent ) {
    /* create a 'Rectangle' region with the input-region bounds of the 'Mask' shape */
    /* the input-region memory is freed and the the 'rectangle' region is returned. */ 
    region = dmi_region_bound( region, fieldx, fieldy );
  } else {
    /* error out */
    if ( region )
      regFree(region);
    region = NULL;
  } 

  return region;
}

/********************************************************************/
/*                                                                  */
/* Function:   dmMaskParse                                          */
/*                                                                  */
/* Description:                                                     */
/*   Parse a mask spec of the form                                  */
/*     sky=mask(file)                                               */
/*                                                                  */
/*   Note: dmRegParse handles 'region' spec                         */
/*                                                                  */
/*  Input parameters:                                               */
/*    char* regstring    Filter region expression for mask          */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
/*  Returns:                                                        */
/*    regRegion *region  Region struct with Mask shape              */
/*                                                                  */
/********************************************************************/
regRegion* dmMaskParse( char* regstring )
{
  dmErrCode  status = dmSUCCESS; 
  char* maskfile;
  dmBlock *block  = NULL;	
  int incl;
  
  if(regstring==NULL)
    return NULL;
    
  /* check and split mask invocation syntax */
  maskfile = dmi_mask_check_syntax( regstring, &incl );
  if(!maskfile)
    return NULL;

  block = dmImageOpen( maskfile );
  
  if (!block){
    if ( fio_access( maskfile ) ) {
      status = DM_ERR_GENERIC;
      sprintf( dm_errbuf, "Mask parsing failure, cannot access external mask file, '%s'.", maskfile );
    } else {
      status = DM_ERR_OPEN_FAIL;
      sprintf( dm_errbuf, "Mask parsing failure, cannot open external mask file, '%s'.", maskfile );
    }    
    dmu_err_set( status, dm_errbuf );
    return NULL;
  }

  regRegion* reg=dmi_mask_read_mask( block );

  free( maskfile );
  dmBlockClose((dmBlock*)block);

  if(reg==NULL)
    return NULL;
  

  regShape *shape=regGetShapeNo(reg, 1);
  if (incl)
    dmi_mask_set_includeflag(shape, regInclude);
  else
    dmi_mask_set_includeflag(shape, regExclude);

  return reg;
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_mask_check_syntax                            */
/*                                                             */
/* Description:                                                */
/*   Check string for appropriate mask invocation syntax.      */
/*   Extracts mask filename from spec.                         */
/*                                                             */
/*   spec can allow                                            */
/*     mask(<maskfile>)     standard invocation                */
/*     -mask(<maskfile>)    exclude syntax                     */
/*     @maskfile            stack syntax?  TBD                 */
/*                                                             */
/* Note:                                                       */
/*   it is illeagel syntax for spec='mask(file)+circle(a,b,r)' */
/*   though the similar syntax works with non-mask shapes      */
/*                                                             */
/*  Input parameters:                                          */
/*    char* spec         filter spec segment                   */
/*                                                             */
/*  Output parameters:                                         */
/*    int*  incl         include/exclude flag                  */
/*                                                             */
/*  Returns:                                                   */
/*    char* filename     mask filename                         */
/*                        NULL if not mask syntax              */
/*                                                             */
/*  Error:                                                     */
/*    On error, sets DM error state                            */
/*                                                             */
/***************************************************************/
char* dmi_mask_check_syntax( char* spec, int *incl )
{
  char *regfile = NULL;

  if ( spec[0] == '@' )
  {
    /* regfile = cs_dup( spec+1 );*/
    sprintf( dm_errbuf, "Mask syntax error, file-directive, '@', is not supported, '%s'.", spec );
    dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
    return regfile;
  }

  /* Check for supported invocation tags */
  long pos_l,pos_r = 0;
  if ( cs_uceq_siz( spec, "-mask(", 6 ) )
  {
    pos_l = 6;
    *incl = 0;  /* regExclude */
  }
  else if ( cs_uceq_siz( spec, "mask(", 5 ) )
  {
    pos_l = 5;
    *incl = 1;  /* regInclude */
  }
  else { /* everything else error */
    pos_l=0;
  }

  /* find the first ")" */
  pos_r=cs_index(spec, ")");
  if( !pos_r               ||   /* if spec has no closing bracket, pos_r is 0 */
      strlen(spec) > pos_r ||   /* if spec is Union: mask()+circle(), the length is beyond the 1st ')' */
      !pos_l               )    /* if 'mask' tag is missing, pos_l=0  */
  {
    sprintf( dm_errbuf, "Mask syntax error, error or unsupported syntax, '%s'", spec );
    dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
    return NULL;
  }

  regfile=cs_dup( spec+pos_l );
  *(regfile+pos_r-pos_l-1)='\0';
  
  /*******************************************-*/
  /*check one more time for wrong syntax for   */
  /* mask(&&circle()                           */
  /*               ^                           */
  /*******************************************-*/
  pos_r=cs_index(regfile, ")");
  if( pos_r != 0 ) {  /* if a compound filter spec has no closing bracket in mask but right one in shape, the pos_r is not 0 */
    sprintf( dm_errbuf, "Mask syntax error, error or unsupported syntax, '%s'", spec );
    dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
    return NULL;
  }

  return regfile;
}

/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_isvalid_mask                                 */
/*                                                                  */
/* Description:                                                     */
/*   Mask block validation for usability.                           */
/*                                                                  */
/*  Input parameters:                                               */
/*    dmBlock* block     Block to examine                           */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
/*  Returns:                                                        */
/*    dmBool             dmTRUE == valid ; dmFALSE == not valid     */
/*                                                                  */
/*                                                                  */
/*   NOTE: This routine is only checking USABILITY, see the sections*/
/*   The full Mask block requirements checking should be done in    */
/*   all 2 levels                                                   */
/*   validation levels:                                             */
/*     level-1: Mask validation,    dmi_mask_isvalid_mask           */
/*     level-2: coords validation,  dmi_mask_isvalid_transf         */
/*                                                                  */
/* Requirement in section 3.1.1, 3.2.1, 4.2, 4.2.2                  */
/*    3.1.1   two dimension                                         */
/*    3.2.1   physical or world coords                              */
/*    4.2     image file                                            */
/*    4.2.2   supported data types                                  */
/*                                                                  */
/********************************************************************/
dmBool dmi_mask_isvalid_mask( dmBlock* block )
{
  dmDescriptor *dd;
  dmDescriptor *dd_phys;
  long naxes;
  long *urpix;

  if(!block)
    return dmFALSE;
  
  /* S4.2,  an image file */
  if( dmi_block_get_raw_type(block )  != dmIMAGE ){
    sprintf( dm_errbuf, "Mask validation failure, mask block '%s' is not an image type.", block->name);
    dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
    return dmFALSE;
  }

  dd = dmImageGetDataDescriptor(block);  
  /* S3.1.1   2-d */
  naxes = dmi_descriptor_get_array_dimensions(dd, &urpix);
  if( naxes != 2){
    if(naxes==0)
      sprintf( dm_errbuf, "Mask validation failure, mask is NULL image.");
    else 
      sprintf( dm_errbuf, "Mask validation failure, mask is not 2 but %ld dimension image.", naxes );
    dmu_err_set( DM_ERR_OPEN_FAIL, dm_errbuf );
    return dmFALSE;
  }

  if ( urpix )
    free(urpix);

  /* 3.2.1  phys and world ?             */
  if( dmArrayGetNoAxisGroups( dd ) != 0) {
    dd_phys = dmArrayGetAxisGroup( dd, 1 ); 

    if ( !dd_phys  ) {
      sprintf( dm_errbuf, "Mask validation failure, no coordinates are defined in mask block '%s'", block->name);
      dmu_err_set( DM_ERR_COORD, dm_errbuf );
      return dmFALSE;
    }    /* 
     else  {
      long ncoords = dmDescriptorGetNoCoords( dd_phys );
      if ( ncoords > 0 ) {
        dmDescriptor *dd_logi;
        dmDescriptor *dd_world;
        // extract next coordinate level (World)    
        dd_world = dmDescriptorGetCoord( dd_phys );
        // extract Parent of the Physical axis set, Logical coordinate set
        dd_logi = dmCoordGetParent( dd_phys );
      }
    }  */
  }
   
  /* 9.2.1.4 subspace   ? */
  //mSubspace *subspace = dmi_block_get_subspace( block ); /* TBD check */

  /* S4.2.2 supported datatypes */
  /* valid datatype - convertable to byte (numeric)   */ 
  char *dtype_str =  dmGetDataTypeName( dd );   
  dmDataType dtype = dmGetDataType(dd);
  switch(dtype){
  case dmDOUBLE:
  case dmFLOAT:
  case dmLONG:
  case dmULONG:
  case dmSHORT:
  case dmUSHORT:
  case dmBYTE:
    break;
  default:
    sprintf( dm_errbuf, "Mask validation failure, invalid datatype ('%s') for mask block '%s'", dtype_str, block->name);
    dmu_err_set(DM_ERR_BAD_TYPE, dm_errbuf );
    return dmFALSE;
  }

  return dmTRUE;
}


/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_get_image_array                              */
/*                                                                  */
/* Description:                                                     */
/*   Extracts the bulk data array from the provided Mask block.     */
/*   Values are returned in thier 'native' data type.               */
/*                                                                  */
/*  Input parameters:                                               */
/*    dmBlock* msk_blk     Mask block                               */
/*                                                                  */
/*  Output parameters:                                              */
/*    long* arraysz      Number of values returned.                 */
/*                                                                  */
/*  Returns:                                                        */
/*    void* values       image values as 1D pixel list.             */
/*                                                                  */
/*  MCD NOTE:                                                       */
/*    wouldn't it be better to return it?                           */
/*    dmi_mask_get_image_array( *block, *dtype, *nvals, *vals )     */
/*    this API may be confusing, most 'get' routines, you tell it   */
/*    which datatype you want rather than what are you getting..    */
/*                                                                  */
/********************************************************************/
void* dmi_mask_get_image_array(dmBlock *msk_blk, dmDataType *dtype, long *arraysz)
{  
  dmDescriptor *dd = dmImageGetDataDescriptor(msk_blk);
  long npix, naxes;
  long status;
  long llpix[2]={1,1};
  long *urpix; 
  void *values;
  
  if( !dd)
    return NULL;

  *dtype = dmGetDataType(dd);
  /* determine image size */
  naxes = dmi_descriptor_get_array_dimensions(dd, &urpix);
  if( !urpix ) {
    sprintf( dm_errbuf, "Mask validation failure, mask image bound array is NULL.");
    dmu_err_set(DM_ERR_GENERIC, dm_errbuf );
    return NULL;
  }

  if(naxes !=2 )
    return NULL;
  npix  = urpix[0] * urpix[1];
  *arraysz=npix;
  switch( *dtype ) {
  case dmBYTE: 
    values = (unsigned char*)malloc(npix*sizeof(unsigned char));
    status = dmp_image_get_subarray(dd, llpix, urpix, dmBYTE, (void*)values);
    break;
  case dmSHORT: 
    values = (short*)malloc(npix*sizeof(short));
    status = dmp_image_get_subarray(dd, llpix, urpix, dmSHORT, (void*)values);
    break;
  case dmUSHORT: 
    values = (unsigned short*)malloc(npix*sizeof(unsigned short));
    status = dmp_image_get_subarray(dd, llpix, urpix, dmUSHORT, (void*)values);
    break;
  case dmLONG: 
    values = (long*)malloc(npix*sizeof(long));
    status = dmp_image_get_subarray(dd, llpix, urpix, dmLONG, (void*)values);
    break;
  case dmULONG: 
    values = (unsigned long*)malloc(npix*sizeof(unsigned long));
    status = dmp_image_get_subarray(dd, llpix, urpix, dmULONG, (void*)values);
   break;
  case dmFLOAT:
    values = (float*)malloc(npix*sizeof(float));
    status = dmp_image_get_subarray(dd, llpix, urpix, dmFLOAT, (void*)values);
    break;
  case dmDOUBLE:
    values = (double*)malloc(npix*sizeof(double));
    status = dmp_image_get_subarray(dd, llpix, urpix, dmDOUBLE, (void*)values);
    break;
  default:
    status = dmFAILURE;
  }
  free(urpix);

  if(status == dmFAILURE){
    sprintf( dm_errbuf, "Mask validation failure, cannot read mask data array.");
    dmu_err_set(DM_ERR_GENERIC, dm_errbuf );
    free(values);
    values=NULL;
  }

  return values;
}

/***************************************************************/
/*                                                             */
/* Function:  dmMaskWrite                                      */
/*                                                             */
/* Description:                                                */
/*   Create MASK block on provided dataset derived from        */
/*   the provided dmIMAGE block.                               */
/*                                                             */
/*                                                             */
/* Requirement in section 3.1.2-3, 4.2.2-4                     */
/*    3.1.2  type Byte                                         */
/*    3.1.3  value 1 or 0                                      */
/*    4.2.2  valid datatypes converted to byte                 */
/*    4.2.3  conversion 0==0, !0==1                            */
/*    4.2.4  IEEE NULL,NaN,Inf -> 0                            */
/*                                                             */
/*  Input parameters:                                          */
/*    dmDataset*  ds     Dataset to get the MASK block         */
/*    char*       name   Name of the MASK block                */
/*    dmBlock*    img    Source 2D Image block                 */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*    dmErrCode   status dmSUCCESS/dmFAILURE                   */
/*                                                             */
/***************************************************************/
dmErrCode dmMaskWrite(dmDataset* ds, char* name, dmMask *newmask)
{
  dmErrCode  status;
  dmDescriptor *dd;
  long  llpix[2]={1,1};
  dmBlock *maskblk;
 
  /* --------------------------------------------- */
  /* create new Mask block in the input dataset    */
  /* --------------------------------------------- */

  maskblk = dmDatasetCreateImage( ds, name, dmBYTE, newmask->axlens, newmask->ndim);
  dd      = dmImageGetDataDescriptor(maskblk);  
  
  
  status=dmSUCCESS;

  if ( maskblk ){
    if( dd ) {
      status = dmi_mask_copy_coordsys(dd, newmask);
      if(status == dmFAILURE){
	sprintf( dm_errbuf, "Mask write failure, cannot create coordinate transformation.");
	dmu_err_set( DM_ERR_GENERIC, dm_errbuf );   
	return status;
      }

      status=dmImageDataSetSubArray_ub(dd, llpix, newmask->axlens, newmask->data);
      if(status == dmFAILURE){
	sprintf( dm_errbuf, "Mask write failure, cannot convert mask image to Byte datatype array.");
	dmu_err_set( DM_ERR_GENERIC, dm_errbuf );   
	return status;
      }
      dmBlockFlush(maskblk);
    } else {
      sprintf( dm_errbuf, "Mask write failure, cannot get mask Byte datatype descriptor.");
      status = dmFAILURE;
    }
  } else {
    sprintf( dm_errbuf, "Mask write failure, cannot create MASK block.");
    status = dmFAILURE;
  }

  if(status == dmFAILURE)
    dmu_err_set( DM_ERR_GENERIC, dm_errbuf );   
  
  return status;
}
/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_get_null                                     */
/*                                                                  */
/* Description:                                                     */
/*  Returns:                                                        */
/*    none                                                          */
/*                                                                  */
/********************************************************************/
void* dmi_mask_get_null(dmDescriptor *dd)
{
  void *nulval;
  dmDataType dtype;

  dtype = dmGetDataType(dd);

  
  switch( dtype ) {      
  case dmBYTE:
    nulval = (unsigned char*)malloc(sizeof(unsigned char));    
    if ( !dmDescriptorGetNull_ub(dd, (unsigned char*)nulval) ){
      free( (unsigned char*)nulval );
      return NULL;
    }
    break;
  case dmSHORT:
    nulval = (short*)malloc(sizeof(short)); 
    if ( !dmDescriptorGetNull_s(dd, (short*)nulval) ){
      free(  (short*)nulval );
      return NULL;
    }
    break;
  case dmUSHORT:
    nulval = (unsigned short*)malloc(sizeof(unsigned short)); 
    if ( !dmDescriptorGetNull_us(dd, (unsigned short*)nulval) ){
      free( (unsigned short*)nulval);
      return NULL;
    }
    break;
  case dmLONG:
    nulval = (long*)malloc(sizeof(long));  
    if ( !dmDescriptorGetNull_l(dd, (long*)nulval) ){
      free((long*)nulval);
      return NULL;
    }
    break;
  case dmULONG:
    nulval = (unsigned long*)malloc(sizeof(unsigned long));      
    if ( !dmDescriptorGetNull_ul(dd, (unsigned long*)nulval) ){
      free((unsigned long*)nulval);
      return NULL;
    }
    break;
  case dmFLOAT:
    nulval =  (float*)malloc(sizeof(float));  
    if ( !dmDescriptorGetNull_f(dd, (float*)nulval) ){
      free((float*)nulval);
      return NULL;
    }
    break;
  case dmDOUBLE:
    nulval = (double*)malloc(sizeof(double)); 
    if ( !dmDescriptorGetNull_d(dd, (double*)nulval) ){
      free((double*)nulval);
      return NULL;
    }
    break;
  default:
    sprintf( dm_errbuf, "Mask write failure, invalid mask image datatype, %d.", dtype);
    dmu_err_set(DM_ERR_BAD_TYPE, dm_errbuf );
    return NULL;
  };
  return nulval;
}

/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_convert_values                               */
/*                                                                  */
/* Description:                                                     */
/*   Convert values from any compatible datatype to byte            */
/*   type, applying rules for MASK value interpretation.            */
/*                                                                  */
/* Requirement in section 3.1.2-3, 4.2.2-4                          */
/*   3.1.2  type Byte                                               */
/*   3.1.3  value 1 or 0                                            */
/*   4.2.2  valid datatypes converted to byte                       */
/*   4.2.3  conversion 0==0, !0==1                                  */
/*   4.2.4  IEEE NULL,NaN,Inf -> 0                                  */
/*                                                                  */
/*  Input parameters:                                               */
/*    void* values       Value array                                */
/*    dmDataType dtype   Datatype of input values.                  */
/*    long  nvals        Number of values in array                  */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
/*  Returns:                                                        */
/*    uc* bytearr        Byte array of Mask values                  */
/*                                                                  */
/********************************************************************/
void* dmi_mask_convert_values( void *values, dmDataType dtype, long nvals, void* nullValue )
{
  long ii;
  unsigned char* vbytes;

  vbytes=(unsigned char*)malloc(nvals*sizeof(unsigned char));
  /* MCD NOTES:                                                   */
  /*   - not sure what ut_q* routines do.. not used within dm     */
  /*   - what is INF for non-floating point?                      */
  /*   - do not see NAN handled.                                  */
  /*   * no value conversion for dmBYTE type..       HHE: added   */
  /*      a byte array may have values 0 -> 255                   */
  /*                                                              */
  
  /* converting .....*/
  /* S4.2.2 - S4.2.4 */
  
  switch( dtype ) {      
  case dmBYTE:
    for (ii=0; ii < nvals; ii++){
      uchar v= ((uchar*)values)[ii]; /* don't have byte-type ut_ func avaiable */
      if( ut_qinf_s((short)v) || ut_qnull_s((short)v) ||  /* S4.2.4  for default INF or NULL */
          (nullValue && v==((uchar*)nullValue)[0])    )   /* S4.2.4, S4.2.3 for user defined NULL */
	vbytes[ii]=0;
      else
	vbytes[ii]= ((uchar*)values)[ii];
    }
    break;
  case dmSHORT:
    for (ii=0; ii < nvals; ii++){
      short v= ((short*)values)[ii];
      if( ut_qinf_s(v) || ut_qnull_s(v) || abs(v)==0 ||  /* S4.2.4 for default INF or NULL  */
	  (nullValue &&  v==((short*)nullValue)[0])    )  /* S4.2.4 for user defined NULL */
	vbytes[ii]=0;
      else 
	vbytes[ii]=1; 
    }
    break;
  case dmUSHORT:
    for (ii=0; ii < nvals; ii++){
      short v= ((unsigned short*)values)[ii];
      if( ut_qinf_s(v) || ut_qnull_s(v)    ||
	  (nullValue &&  v==((unsigned short*)nullValue)[0]) )   /* S4.2.4 */
	vbytes[ii]=0;
      else 
	vbytes[ii]=1; 
    }
    break;
  case dmLONG:
    for (ii=0; ii < nvals; ii++){
      long v= ((long*)values)[ii];
      if( ut_qinf_i(v) || ut_qnull_i(v) || labs(v)==0 || 
	  (nullValue &&  v==((long*)nullValue)[0] )      )  /* S4.2.4 */
	vbytes[ii]=0;
      else 
	vbytes[ii]=1; 
    }
    break;
  case dmULONG:
    for (ii=0; ii < nvals; ii++){
      long v= ((unsigned long*)values)[ii];
      if( ut_qinf_i(v) || ut_qnull_i(v)  || 
	  (nullValue &&  v==((unsigned long*)nullValue)[0] )  )  /* S4.2.4 */
	vbytes[ii]=0;
      else 
	vbytes[ii]=1; 
    }
    break;
  case dmFLOAT:
    for (ii=0; ii < nvals; ii++){
      float v = ((float*)values)[ii];
      if( ut_qinf_r(v) || ut_qnull_r(v) || ut_qnan_r(v) || /* S4.2.4  for default INF or NULL or NAN */
	  (nullValue &&  fabsf(v-((float*)nullValue)[0])<FLT_MIN ) ||  /* S4.2.4  for user defined  INF or NULL */
	  fabsf(v) < FLT_MIN ) {  /* S4.2.3  */
    	vbytes[ii]=0;
      }  else
	vbytes[ii]=1; 
    }
    break;
  case dmDOUBLE:
    for (ii=0; ii < nvals; ii++){
      double v= ((double*)values)[ii];
      if( ut_qinf_d(v) || ut_qnull_d(v) || ut_qnan_d(v) || /* S4.2.4 for default INF or NULL or NAN */
	  (nullValue &&  fabs(v-((double*)nullValue)[0])<DBL_MIN ) ||  /* S4.2.4  for user defined  INF or NULL */
	  fabs(v) < DBL_MIN ) {  /* S4.2.3  */
	vbytes[ii]=0;
      }  else
	vbytes[ii]=1; 
    }
    break;
  default:
    sprintf( dm_errbuf, "Mask write failure, invalid mask image datatype, %d.", dtype);
    dmu_err_set(DM_ERR_BAD_TYPE, dm_errbuf );
    free(vbytes);
    return NULL;
  }; /* end of convertion */


  return vbytes;
}
 
/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_read_mask                                    */
/*                                                                  */
/* Description:                                                     */
/*   Validate and create Mask region/shape                          */
/*                                                                  */
/*  Input parameters:                                               */
/*    dmBlock  *block    Image block for use as Mask                */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
/*  Returns:                                                        */
/*    regRegion *region                                             */
/*                                                                  */
/********************************************************************/
regRegion* dmi_mask_read_mask( dmBlock* block )
{ 
  regRegion *myRegion = NULL;
  regShape  *myShape;
  dmMask    *smask;
  long ii;

  /* validate the block is appropraite Mask image block*/
  if( !dmi_mask_isvalid_mask(block) )
    return NULL;

  /* create MASK region shape attributes */ 
  myShape             = ( regShape *) calloc ( 1, sizeof( regShape ) );
  myShape->type       = regMASK;
  myShape->include    = regInclude;       /* updated somewhere else */

  /* xpos and ypos cache mask extent.. values assigned later      */
  /* right now, we do not know which coordinate system to use.  */
  myShape->nPoints   = 2;
  myShape->xpos      = calloc( 2, sizeof(double) );
  myShape->ypos      = calloc( 2, sizeof(double) );
  myShape->flag_coord= 0; 
  //  myShape->component  =1; TBEvaluated

  // zeroing
  for(ii=0; ii<2;ii++){
    myShape->xpos[ii]=0.0;
    myShape->ypos[ii]=0.0;
  }
  dmDescriptor *dd0, *dd;
  dd0 =  dmImageGetDataDescriptor(block); // rtn  block->cols[0]
  // what if (dmBlockGetType(dd->block) == dmIMAGE 
  // dd0->ngroups>1
  // for vector, sky,    dd0->ngroups=1
  // for list, (pi,pha), dd0->ngroups=2
  long ngrp = dmi_descriptor_get_no_groups(dd0);
  if( ngrp==1 ) {
    dd  = dmArrayGetAxisGroup( dd0, 1 );    /* physical coordinates */
    // default to PHYSICAL Coords 
    if (dd->flag & DM_FLAG_PHYS ) 
      ; //okay 
    else    if ( dd->flag & DM_FLAG_LOGICAL ) 
      ; //okay 
    else {
      // not supported WORLD coords at this time 
      sprintf( dm_errbuf, "ERROR: Mask coordinate, 'World', is not supported.");
      dmu_err_set( DM_ERR_PARSE, dm_errbuf );
      return  NULL;
    }
    
  } else if( ngrp==2 ){
    for(ii=1; ii<=ngrp; ii++){
      dd=dmi_descriptor_get_group_no(dd0,ii);
      if (dd){
	// default to PHYSICAL Coords 
	if ( (dd->flag & DM_FLAG_LOGICAL) || (dd->flag & DM_FLAG_PHYS) ) 
	  ; //okay 
	else {
	  // not supported WORLD coords at this time 
	  sprintf( dm_errbuf, "ERROR: Mask coordinate, 'World', is not supported.");
	  dmu_err_set( DM_ERR_PARSE, dm_errbuf );
	  return  NULL;
	}
      } else {
	sprintf( dm_errbuf, "ERROR: Mask coordinate, failed to read axis-group descriptor.");
	dmu_err_set( DM_ERR_PARSE, dm_errbuf );
      return  NULL;      
      }
    }
  }
    
  //set region-shape coordinate flag
  if (dd->flag & DM_FLAG_PHYS)
    myShape->flag_coord = RC_PHYSICAL; 
  else if (dd->flag & DM_FLAG_LOGICAL)
    myShape->flag_coord = RC_LOGICAL;

  myShape->isInside  = dmi_mask_inside;
  myShape->copy      = dmi_mask_copy;
  myShape->isEqual   = dmi_mask_isequal;
  myShape->toString  = dmi_mask_tostring;
  myShape->calcExtent= dmi_mask_extent;
  myShape->calcArea  = dmi_mask_area;

  myShape->name = dmMaskName();
  
  smask = (dmMask*)dmi_mask_convert(block);
  if( smask == NULL)
    return NULL;
  myShape->spec = smask;
 
  // assign shape->xpos,ypos values with mask-struct-bounds
  myShape->xpos[0] =  smask->xbounds[0];
  myShape->xpos[1] =  smask->xbounds[1];
  myShape->ypos[0] =  smask->ybounds[0];
  myShape->ypos[1] =  smask->ybounds[1];

  myRegion  = regCreateRegion(NULL,NULL);
  regAddShape(myRegion,1,myShape);

  return myRegion; 
}   

/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_set_includeflag                              */
/*                                                                  */
/* Description:                                                     */
/*   Set include flag attribute of Mask Shape                       */
/*                                                                  */
/*  Input parameters:                                               */
/*    regShape* shape    Mask Shape                                 */
/*    int       fincl    Include flag                               */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
/*  Returns:                                                        */
/*    none                                                          */
/*                                                                  */
/********************************************************************/
void dmi_mask_set_includeflag(regShape *shape, int fincl)
{
  if ( shape != NULL )
    shape->include = fincl;
}

/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_free                                         */
/*                                                                  */
/* Description:                                                     */
/*   Bridge function for regUSER shape.                             */
/*   Free memory allocated to members of struct regUSERSHAPE        */
/*                                                                  */
/*  Input parameters:                                               */
/*    regShape shape     Mask region shape                          */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
/*  Returns:                                                        */
/*    none                                                          */
/*                                                                  */
/********************************************************************/
void dmi_mask_free(regShape *shape)
{

  regFreeShape(NULL, shape);

}

/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_tostring                                     */
/*                                                                  */
/* Description:                                                     */
/*   Bridge function for regUSER shape.                             */
/*   Generates string representation of the shape.                  */
/*     Spec: 10.1.1.1                                               */
/*                                                                  */
/* Note:                                                            */
/*   User must free the returned memory.                            */
/*                                                                  */
/*  Input parameters:                                               */
/*    regShape *shape    Mask region shape                          */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
/*  Returns:                                                        */
/*    char*  buff        String rep of Mask shape.                  */
/*                                                                  */
/********************************************************************/
void dmi_mask_tostring(regShape *shape, char* sbuf, long size)
{
  char blkname[TEXT_SIZE];
  dmMask *smask;
  
  smask = dmi_mask_get_spec(shape);
  strcpy(blkname, smask->blkname);
  if( strlen(blkname)==0)
    strcpy(blkname, "UNKNOWN");

  /* allocate for format "<shapename>(<blockname>)\0" */    
  snprintf(sbuf, size, "%s(%s)", shape->name, blkname);

}

/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_area                                         */
/*                                                                  */
/* Description:                                                     */
/*   Bridge function for regUSER shape.                             */
/*   Determines area of Mask shape.                                 */
/*                                                                  */
/*  Input parameters:                                               */
/*    regShape shape     Mask region shape                          */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
/*  Returns:                                                        */
/*    double area        Area of Mask shape                         */
/*                                                                  */
/********************************************************************/
double dmi_mask_area( regShape *shape )
{
  long npix, ii, count;
  double area=0.0;
  unsigned char *vbytes;
  dmMask *smask;

  smask = dmi_mask_get_spec(shape);
  /* extract pixel data array */
  vbytes = smask->data;
  if(vbytes == NULL)
    return area;

  /* count 'on' pixels (value==1) */
  count = 0; 
  npix = smask->axlens[0]*smask->axlens[1];
  for (ii=0; ii < npix; ii++)
    if ( vbytes[ii] == 1 ) count++; 

  area = (double)count;  
  if ( ((dmMask*)shape->spec)->cflag & DM_FLAG_PHYS ) 
    /* scale the counts by pixel size */
    area *=  smask->log2phys->delt[0] *  smask->log2phys->delt[1];    
 

  return area;
}

/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_maskstruct_invert_bounds                     */
/*                                                                  */
/* Description:                                                     */
/*   Transform the current coordinates from logical to physical     */
/*   or vise versa.                                                 */
/*                                                                  */
/*  Input parameters:                                               */
/*    dmMask* smask     Mask struct pointer                         */
/*                                                                  */
/*  Output parameters:                                              */
/*                                                                  */
/*  Returns:                                                        */
/*    int       error status                                        */
/*                                                                  */
/*                                                                  */
/********************************************************************/
int dmi_mask_maskstruct_invert_bounds(dmMask *smask, int invert)
{ 
  double inpos[2];
  double outpos[2];
  int dim;

  // invert smask's bounds from PHYS to LOGICAL or vise versa
  if( !smask || !smask->xbounds || !smask->ybounds || !smask->log2phys )
    return dmFAILURE;

  dim      = smask->ndim;

  inpos[0] = smask->xbounds[0];
  inpos[1] = smask->ybounds[0];
  dmi_transform_apply( smask->log2phys, &inpos[0], &outpos[0], dim, invert); 
  smask->xbounds[0] = outpos[0];
  smask->ybounds[0] = outpos[1];
  
  inpos[0] = smask->xbounds[1];
  inpos[1] = smask->ybounds[1];
  dmi_transform_apply( smask->log2phys, &inpos[0], &outpos[0], dim, invert); 
  smask->xbounds[1] = outpos[0];
  smask->ybounds[1] = outpos[1];

  return dmSUCCESS;
}

/********************************************************************/
/*                                                                  */
/* Function: dmi_mask_extent                                        */
/*                                                                  */
/* Description:                                                     */
/*   Bridge function for regUSER shape.                             */
/*   Determines the bounds of the Mask shape                        */
/*                                                                  */
/*  Input parameters:                                               */
/*    regShape* shape    Mask region shape                          */
/*                                                                  */
/*  Output parameters:                                              */
/*    double*  xpos      X-axis bounds [min,max]                    */
/*    double*  ypos      Y-axis bounds [min,max]                    */
/*                                                                  */
/*  Returns:                                                        */
/*    int        ok      Error flag; 1-success, 0-problem           */
/*                                                                  */
/********************************************************************/
int dmi_mask_extent( regShape *shape, double *xpos, double *ypos )
{
  dmMask *smask;
  smask = (dmMask*)dmi_mask_get_spec(shape);


  // take it coordinate reference  shape->user->flag_coord 
  if ( (shape->flag_coord == RC_LOGICAL) && (smask->cflag & DM_FLAG_PHYS) ){    
    dmi_mask_maskstruct_invert_bounds(smask, dmTRUE);
    //update the smask's cflag
    smask->cflag = DM_FLAG_LOGICAL;

  } else if ( (shape->flag_coord == RC_PHYSICAL) && (smask->cflag & DM_FLAG_LOGICAL) ){    
    dmi_mask_maskstruct_invert_bounds(smask, dmFALSE);
    //update the smask's cflag
    smask->cflag = DM_FLAG_PHYS;
  } 

  
  /* smask bounds are on coords indicated by shape->user->flag_coor */
  xpos[0] = smask->xbounds[0];
  xpos[1] = smask->xbounds[1];
  ypos[0] = smask->ybounds[0];
  ypos[1] = smask->ybounds[1];


  // true on flag_coord == DM_FLAG_LOGICAL
  if ( shape->flag_coord == RC_LOGICAL ){    
    if( shape->xpos && shape->ypos){
      xpos[0] += shape->xpos[0];  /* shape->xpos is the pos relative to the ll and up coners of Event Image frame */
      ypos[0] += shape->ypos[0];  /* shape->ypos is the pos relative to the ll and up coners of Event Image frame */
      xpos[1] += shape->xpos[0];  /* which are calculated in region_create.c:regConvertWorldShape() */
      ypos[1] += shape->ypos[0];

    }
  } 

  if(dmverbose>=5){
    printf("Mask image's bounds at Event ");
    if ( shape->flag_coord == RC_LOGICAL )
      printf("\nlogical coords:");
    else
      printf("\nphysical coords:");
    
    printf("\n  xpos=(%10.3f, %10.3f) ypos=(%10.3f,%10.3f) (PIXL)\n", xpos[0],xpos[1],ypos[0],ypos[1]);
  }

  return 1;
}

/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_isequal                                      */
/*                                                                  */
/* Description:                                                     */
/*   Bridge function for regUSER shape.                             */
/*   Checks content of Mask Shapes for equality.                    */
/*                                                                  */
/*  Input parameters:                                               */
/*    regShape shape1    Mask region shape                          */
/*    regShape shape2    Mask region shape                          */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
/*  Returns:                                                        */
/*    int result         dmTRUE(1)  = Masks are equal               */
/*                       dmFALSE(0) = Masks are not equal           */
/*                                                                  */
/********************************************************************/
int dmi_mask_isequal(regShape *shape1, regShape *shape2)
{
  long ii;
  dmMask* smask1;
  dmMask* smask2;

  /*                                  */
  /*  check mask-shape attributes     */
  /*                                  */
  if(shape1->type != shape2->type               ||
     shape1->include != shape2->include         ||
     shape1->nPoints != shape2->nPoints         ||
     shape1->flag_coord != shape2->flag_coord    ||
     //     myShape->component   != inshp->component   || TBEvaluated
     !cs_eq(shape1->name,shape2->name) 
     )
    return dmFALSE;

  if(shape1->nPoints == 0 )
    return dmFALSE;
  if( !shape1->xpos || !shape2->xpos || 
      !shape1->ypos || !shape2->ypos )
    return dmFALSE;
  for(ii=0;ii< shape1->nPoints;ii++){
    if(shape1->xpos[ii] != shape2->xpos[ii] ||
       shape1->ypos[ii] != shape2->ypos[ii])
      return dmFALSE;
  }     
  

  /*                                  */
  /* check  mask block image contents */
  /*                                  */
  smask1 = dmi_mask_get_spec(shape1);
  smask2 = dmi_mask_get_spec(shape2);

  if(!smask1 || !smask2)
    return dmFALSE;
 
  /* check  masks equal or not  */
  return dmi_mask_maskstruct_isequal(smask1,smask2);

}

/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_isequal_maskstruct                           */
/*                                                                  */
/* Description:                                                     */
/*   compare two mask structs if they are completely identicl       */
/*   in all aatributes and contents.                                */
/*                                                                  */
/*  Input parameters:                                               */
/*    dmMask *smask1    Mask struct pointer                         */
/*    dmMask *smask2    Mask struct pointer                         */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
/*  Returns:                                                        */
/*    dmBool        dmTRUE(1)  = Mask-structs are equal             */
/*                  dmFALSE(0) = Mask-structs are not equal         */
/*                                                                  */
/********************************************************************/
dmBool dmi_mask_maskstruct_isequal(dmMask *smask1, dmMask* smask2)
{
  int ii;
  
  /* check  mask block name  */
  if( !cs_eq(smask1->blkname, smask1->blkname) ) 
    return dmFALSE;
    
  /* check  image bounds and size */
  if(smask1->ndim != smask2->ndim )
    return dmFALSE;
  for (ii=0; ii< smask1->ndim; ii++ ){
    if( smask1->axlens[ii]  != smask2->axlens[ii]  ||
	smask1->xbounds[ii] != smask2->xbounds[ii] || 
	smask1->ybounds[ii] != smask2->ybounds[ii] )
      return dmFALSE;
  }

   /* check  image  coords */
  if( smask1->log2phys && smask2->log2phys) {
    if( !cs_eq(smask1->log2phys->name, smask2->log2phys->name)     ||
	!cs_eq(smask1->log2phys->type, smask2->log2phys->type)     ||
	smask1->log2phys->physical != smask2->log2phys->physical )
      return dmFALSE;
    for (ii=0; ii< smask1->ndim; ii++ ){
      if( smask1->log2phys->val[ii]  != smask2->log2phys->val[ii] ||
	  smask1->log2phys->pix[ii]  != smask2->log2phys->pix[ii] ||
	  smask1->log2phys->delt[ii] != smask2->log2phys->delt[ii] )
	return dmFALSE;
    }
  } else if( !smask1->log2phys && smask2->log2phys ) 
    return dmFALSE;
  else if( smask1->log2phys && !smask2->log2phys ) 
    return dmFALSE;

  if( smask1->phys2wcs && smask2->phys2wcs) {
    if( !cs_eq(smask1->phys2wcs->name, smask2->phys2wcs->name)    ||
	!cs_eq(smask1->phys2wcs->type, smask2->phys2wcs->type)    ||
	!cs_eq(smask1->phys2wcs->radesys, smask2->phys2wcs->radesys)    ||
	smask1->phys2wcs->theta    != smask2->phys2wcs->theta    ||
	smask1->phys2wcs->longpole != smask2->phys2wcs->longpole ||
	smask1->phys2wcs->latpole  != smask2->phys2wcs->latpole  ||
	smask1->phys2wcs->equinox  != smask2->phys2wcs->equinox  )
      return dmFALSE;
 
    for (ii=0; ii< smask1->ndim; ii++ ){
      if( smask1->phys2wcs->val[ii]  != smask2->phys2wcs->val[ii] ||
	  smask1->phys2wcs->pix[ii]  != smask2->phys2wcs->pix[ii] ||
	  smask1->phys2wcs->delt[ii] != smask2->phys2wcs->delt[ii] )
	return dmFALSE;
    }
  } else if( !smask1->phys2wcs && smask2->phys2wcs ) 
    return dmFALSE;
  else if( smask1->phys2wcs && !smask2->phys2wcs ) 
    return dmFALSE;
   

  /* check  image  data */
  long npix=smask1->axlens[0] * smask1->axlens[1];
  for(ii=0;ii<npix;ii++){
    if(smask1->data[ii] != smask2->data[ii] )
      return dmFALSE;
  }
  
  return dmTRUE;
}
/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_copy                                         */
/*                                                                  */
/* Description:                                                     */
/*   Bridge function for regUSER shape.                             */
/*   Make a copy of the provide Mask Shape                          */
/*                                                                  */
/*  Input parameters:                                               */
/*    regShape shape     Mask region shape to copy                  */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
/*  Returns:                                                        */
/*    regShape shape     resulting Mask shape                       */
/*                                                                  */
/*   NOTE:                                                          */
/*     component, region, next attributes are NOT copied            */
/*                                                                  */
/********************************************************************/
regShape* dmi_mask_copy(regShape *inshp)
{
  regShape *myShape;
  dmMask *smask;
  long ii;

 
  myShape = ( regShape *) calloc ( 1, sizeof( regShape ) );

  myShape->type        = inshp->type;
  myShape->include     = inshp->include;
  myShape->nPoints     = inshp->nPoints;
  myShape->flag_coord  = inshp->flag_coord; 
  myShape->component   = inshp->component;

  if( inshp->xpos && inshp->ypos && (inshp->nPoints > 0) )
  {
    myShape->xpos=(double *)calloc( myShape->nPoints, sizeof(double));
    myShape->ypos=(double *)calloc( myShape->nPoints, sizeof(double));

    for(ii=0;ii<inshp->nPoints;ii++){
      myShape->xpos[ii]=inshp->xpos[ii];
      myShape->ypos[ii]=inshp->ypos[ii];
    }
  }
 
  myShape->name            = cs_dup( inshp->name );
  myShape->isInside        = inshp->isInside;
  myShape->copy            = inshp->copy;
  myShape->isEqual   = inshp->isEqual;
  myShape->toString  = inshp->toString;
  myShape->calcExtent= inshp->calcExtent;
  myShape->calcArea  = inshp->calcArea;

  
  /* get new mask-struct pointer and contents  */
  smask = dmi_mask_get_spec( inshp);
  myShape->spec = (dmMask*)dmi_mask_maskstruct_copy(smask);
  
  return myShape; 
}

/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_isvalid_transf                               */
/*                                                                  */
/* Description:                                                     */
/*   Validated Mask DM-descriptor coords transformation matching    */
/*             filter DM-descriptor coords transformation           */
/*                                                                  */
/*  Input parameters:                                               */
/*    regShape* shape    Mask region shape pointer                  */
/*    dmBlock* src_blk   Source dm-block pointer                    */
/*                                                                  */
/*  Output parameters:                                              */
/*                                                                  */
/*  Returns:                                                        */
/*    dmBool            TRUE if coords matching and FALSE otherwise */
/*                                                                  */
/*                                                                  */
/* Requirement in section 4.3.3                                     */
/*   4.3.3  projection type (not reference point)                   */
/*                                                                  */
/* Linear Coords transforming to X given x:                         */ 
/*    X = a + b * (x-c)                                             */
/* a: offset  b: scale   c: reference                               */
/* or                                                               */
/*    X = A + bx    where  A = a - b*c                              */
/*                                                                  */
/* To validate a match of two linear coordinates, say               */
/*    X  = A  + bx    where  A  = a  - b*c                          */
/*    X' = A' + b'x   where  A' = a' - b'*c'                        */
/*  and an exact match in pixel by pixel (with a simpler way)       */
/*  two conditions must be met                                      */  
/*    b == b'             (same scale)                              */
/*    A-A' == integer     (pixel boundaries align)                  */
/*                                                                  */
/* NOTE:                                                            */
/*   Validation of source transform should not happen for 'regular' */
/*   Table data (ie: NOT defined by a transform).                   */
/*   In the DM; this is represented by a Unit transform on the      */
/*   coordinate descriptor.                                         */
/*                                                                  */
/********************************************************************/
dmBool dmi_mask_isvalid_transf(regShape *shape, dmDescriptor *dd)
{
 
 /* ------------------------------------------------------*- */
  /* retrieve descriptors and coords of Mask and filter       */
  /* validate the same coords of Mask and filter descriptor   */
  /* validate coords transformation paramters, if not LOGICAL */
  /* realize Requirement section 4.3.3                        */
  /* ------------------------------------------------------*- */    

  if ( !dd ) {
    sprintf( dm_errbuf, "Mask validation failure, cannot find source block column.");
    dmu_err_set( DM_ERR_PARSE, dm_errbuf );
    return  dmFALSE;
  }

  if( (dmBlockGetType(dd->block) == dmTABLE ) && dmi_coord_transform_is_unit(dd) ) {
    /* Table column with Unit transform == regular.  No validation     */
    /* of transform definitions should/need be performed.              */
      return dmTRUE;
  }

  if ( (((dmMask*)shape->spec)->cflag & DM_FLAG_LOGICAL) &&  (dd->flag & DM_FLAG_LOGICAL) ) {
    // nothing to be validated further
    return dmTRUE;
  }
 
  /* ----------------------------------------------------------------- */
  /* validate coords transformation paramters, if not LOGICAL          */
  /* between Mask and filter descriptors                               */
  /*                                                                   */
  /* Linear Coords transforming to X given x:                          */ 
  /* X = a + b * (x-c)                                                 */
  /* a: offset  b: scale   c: reference                                */
  /* or                                                                */
  /* X = A + bx    where  A = a - b*c                                  */
  /*                                                                   */
  /* To validate a match of two linear coordinates, say                */
  /*  X  = A  + bx    where  A = a - b*c                               */
  /*  X' = A' + b'x   where  A' = a' - b'*c'                           */
  /*  and an exact match in pixel by pixel (with a simpler way)        */
  /*  two conditions must be met                                       */  
  /*   b == b'                                                         */
  /*   A-A' == integer                                                 */
  /* ----------------------------------------------------------------- */    
  int ii;
  float diff_val, diff_dlt, diff_pix;
  float torl=1.e-5;   //FLT_EPSILON    TBdecided
  dmMask *smask;
 
  dmCoordTransform* log2phys_1=NULL;
  dmCoordTransform* phys2wcs_1=NULL;
  dmCoordTransform* log2phys_2=NULL;
  dmCoordTransform* phys2wcs_2=NULL;

  smask = (dmMask*)dmi_mask_get_spec( shape);

  /* get coords transformations */
  log2phys_1 = smask->log2phys;
  phys2wcs_1 = smask->phys2wcs;
  dmi_mask_get_transforms( dd, &log2phys_2, &phys2wcs_2);
 
  /*******************************************************************************************
   validate the same coords transformation type between Mask and event/filter descriptors 
  *******************************************************************************************/
  if( log2phys_1 && log2phys_2){
    if( !cs_uceq( log2phys_1->type, log2phys_2->type ) ) {
      sprintf( dm_errbuf, "Mask coordinate failure, mismatch transformation type: '%s' vs '%s'.", log2phys_1->type, log2phys_2->type);
      dmu_err_set( DM_ERR_COORD, dm_errbuf );
      return  dmFALSE;
    } 

    /* look up the vector name string, looser match */
    if ( !cs_uceq(log2phys_1->name, log2phys_2->name )){
      if(dmverbose>1)
	printf("\nMask validation WARN, mismatch descriptor name, Mask's '%s' vs Target's '%s',\n",  log2phys_1->name, log2phys_2->name);
    }

  } else {
    sprintf( dm_errbuf, "Mask coordinate failure, linear ransformation is NULL");
    dmu_err_set( DM_ERR_COORD, dm_errbuf );
    return  dmFALSE;
  }

  if( phys2wcs_1 &&  phys2wcs_2 ){
    if( !cs_uceq( phys2wcs_1->type, phys2wcs_2->type ) ) {
      sprintf( dm_errbuf, "Mask coordinate failure, mismatch projection type: '%s' vs '%s'.", phys2wcs_1->type, phys2wcs_2->type);
      dmu_err_set( DM_ERR_COORD, dm_errbuf );
      return  dmFALSE;
    }

    /* look up the vector name string, looser match */
    if ( !cs_uceq(phys2wcs_1->name, phys2wcs_2->name )){
      if(dmverbose>1)
	printf("\nMask validation WARN, mismatch descriptor name, Mask's '%s' vs Target's '%s',\n",  phys2wcs_1->name, phys2wcs_2->name);
    }
  } else if( (phys2wcs_1 && !phys2wcs_2) ||  (!phys2wcs_1 && phys2wcs_2) ){
    if( phys2wcs_1 && !phys2wcs_2)
      sprintf( dm_errbuf, "Mask coordinate failure, mismatch '%s' transformation.", phys2wcs_1->name);
    else
      sprintf( dm_errbuf, "Mask coordinate failure, mismatch '%s' transformation.", phys2wcs_2->name);
    dmu_err_set( DM_ERR_COORD, dm_errbuf );
    return  dmFALSE;
  }


 /*******************************************************************************************
   validate the same coords transformation params between Mask and event/filter descriptors 
    
    y = offset + scale * ( x - reference )
    val[2]: offset
   delt[2]: scale     
    pix[2]: reference
   *******************************************************************************************/

  // if scale is different, take the ordinary way, checking offset and reference individually
  dmBool scaleSame=dmTRUE;
  for( ii = 0; ii < smask->ndim; ii++ ) {
    if( log2phys_1->delt[ii] !=  log2phys_2->delt[ii] )
      scaleSame=dmFALSE;
  }

  double A_1, A_2;
  for( ii = 0; ii < smask->ndim; ii++ ) {
    if (  !scaleSame ) {
      // check offset and references individually
      diff_val = fabs(log2phys_1->val[ii] - log2phys_2->val[ii]);
      diff_pix = fabs(log2phys_1->pix[ii] - log2phys_2->pix[ii]);
      
      if( diff_val >= torl  ){ 
	sprintf( dm_errbuf, "Mask coordinate failure, mismatch '%s' offset for '%s'.",log2phys_1->type, log2phys_1->name);
	dmu_err_set( DM_ERR_COORD, dm_errbuf );
	return  dmFALSE;
      } else if( diff_pix >= torl  ){ 
	sprintf( dm_errbuf, "Mask coordinate failure, mismatch '%s' reference for '%s'.",log2phys_1->type, log2phys_1->name);
	dmu_err_set( DM_ERR_COORD, dm_errbuf );
	return  dmFALSE;
      }
    } else {
      // check offset and references combinely
      A_1 = log2phys_1->val[ii] - log2phys_1->delt[ii] * log2phys_1->pix[ii];
      A_2 = log2phys_2->val[ii] - log2phys_2->delt[ii] * log2phys_2->pix[ii];
      
      diff_pix = fabs(  (A_1-A_2) - (int)(A_1-A_2) );
      if( diff_pix >= torl  ){ 
	sprintf( dm_errbuf, "Mask coordinate failure, mismatch '%s' offset+ref for '%s'.",log2phys_1->type, log2phys_1->name);
	dmu_err_set( DM_ERR_COORD, dm_errbuf );
	return  dmFALSE;
      }
    }

    if( phys2wcs_1 &&  phys2wcs_2 ) {
      /* for TAN tyep: ignore checking the reference  */
      diff_val = fabs(phys2wcs_1->val[ii]  - phys2wcs_2->val[ii]);
      diff_dlt = fabs(phys2wcs_1->delt[ii] - phys2wcs_2->delt[ii]);
      diff_pix = fabs(phys2wcs_1->pix[ii]  - phys2wcs_2->pix[ii]);
      if( diff_val >= torl || diff_dlt >= FLT_EPSILON  ) {
	if( diff_val >= torl )
	  sprintf( dm_errbuf, "Mask coordinate failure, mismatch '%s' transformation offsets for '%s'.", phys2wcs_1->type, phys2wcs_1->name);
	else
	  sprintf( dm_errbuf, "Mask coordinate failure, mismatch '%s' transformation deltas for '%s'.", phys2wcs_1->type, phys2wcs_1->name);

	dmu_err_set( DM_ERR_COORD, dm_errbuf );
	return  dmFALSE;
      }
    } // end of  if( phys2wcs_1 &&  phys2wcs_2 )

  } // end of for( ii = 0; ii < dim; ii++ ) 


  if(dmverbose>=5){
    dmi_mask_print_transform(log2phys_1);
    dmi_mask_print_transform(phys2wcs_1);
    printf("\n\n\tvs\n");
    dmi_mask_print_transform(log2phys_2);
    dmi_mask_print_transform(phys2wcs_2);
    printf("\n");
  }

  // return diffs of the coords in scale factors
  if( !scaleSame )
    return (dmTRUE+dmFALSE+1);
  else 
    return dmTRUE;
}

/********************************************************************/
/*                                                                  */
/* Function:   dmi_mask_get_spec                                    */
/*                                                                  */
/* Description:                                                     */
/*   return a Mask struct pointer from region shape                 */
/*                                                                  */
/*  Input parameters:                                               */
/*    regShape*  shape   Mask Region shape struct                   */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
/*  Returns:                                                        */
/*    dmMask* smask   Mask-struct from MASK shape.                  */
/*                                                                  */
/********************************************************************/
void *dmi_mask_get_spec(regShape *myShape)
{
  return myShape->spec;
}

/********************************************************************/
/*                                                                  */
/* Function:   dmi_mask_set_spec                                    */
/*                                                                  */
/* Description:                                                     */
/*   assign Mask struct pointer to region shape                     */
/*                                                                  */
/*  Input parameters:                                               */
/*    regShape*  shape   Mask Region shape                          */
/*    void *             Mask struct pointer                        */
//*                                                                 */
/*  Output parameters:                                              */
/*    update shape with Mask struct pointernone                     */
/*                                                                  */
/*  Returns:                                                        */
/*    none                                                          */
/*                                                                  */
/********************************************************************/
void  dmi_mask_set_spec(regShape* shape, void* smask)
{ 
  shape->spec=(dmMask*)smask;
}

/****************************************************************** */
/*                                                                  */
/* Function:  dmi_mask_match_coords                                 */
/*                                                                  */
/* Description:                                                     */
/*                                                                  */
/*  Validate Mask vector component names matching to event-list's   */
/*  Validate Mask coords matching to event-list's, including        */
/*  offset and reference but not scale which will be checked in     */
/*  mask rebinning section later.                                   */
/*                                                                  */
/* The *strict* approach would be to require both the vector-name   */
/* and the component-names to match. A slightly looser approach     */
/* would be to require the component names to match but allow       */
/* different vector-names (e.g. allow sky(a,b) to match pos(a,b) ). */
/* Do not allow sky(a,b) to match sky(p,q) and error out.           */
/*                                                                  */
/*                                                                  */
/*  Input parameters:                                               */
/*    regShape* myShape   Mask region shape                         */
/*    dmBlock* src_blk    source block pointer                      */
/*    char* cptname       Name of filter descriptor:                */
/*                        either vector name string                 */
/*                        or components name string                 */     
/*                                                                  */
/*  Output parameters:                                              */
/*                     None                                         */
/*                                                                  */
/*  Returns:                                                        */
/*    dmBool            TRUE if coords matching or FALSE if not     */
/*                                                                  */
/* Requirement in section 4.3.3                                     */
/*   4.3.3  projection type and (not reference point)               */
/*                                                                  */
/****************************************************************** */
dmBool dmi_mask_match_coords(regShape *myShape,    dmBlock *src_blk, char *cptname)
{
  int ii       ;
  dmMask *smask;
  dmDescriptor *dd0, *dd;
  dmBool stat;

  smask = (dmMask*)dmi_mask_get_spec(myShape);
 
  // get the dmDescriptor from source dmBlock given the component-name string
  dd0 = dmi_block_search_raw_col_identifier( src_blk, cptname );
  dd  = dd0;
  if( !dd ){
    sprintf( dm_errbuf, "Mask validation failure, invalid source column descriptor for '%s'",  cptname);
    dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
    return dmFALSE;
  }

  // look up the dimensions  between the mask's and src's
  if( smask->ndim != dd->dim ){
    sprintf( dm_errbuf, "Mask validation failure, mismatch dimensions of mask and source: %ld vs %ld.",  smask->ndim,dd->dim);
    dmu_err_set( DM_ERR_GENERIC, dm_errbuf );    
    return dmFALSE;
  }

  // look up dmCoordTransform attributes:
  // stat==dmTRUE:  matched
  // stat==dmFALSE: not matched in ref or offset
  // stat==<else>:  not matched in scale, to be checked if it is a valid multiple later
  stat = dmi_mask_isvalid_transf(myShape, dd);
  if(stat == dmFALSE)
    return dmFALSE;
  
  // look up the componens-name string of the vector, exact match 
  if( dd->flag & DM_FLAG_LOGICAL ) 
    dd = dmDescriptorGetCoord( dd );
  for(ii=0; ii<smask->ndim; ii++){
    if ( !dd->cpts[ii] && smask->l_cptNames[ii] ){
      sprintf( dm_errbuf, "Mask validation failure, mismatch descriptor component name of mask and source: \"%s(%s,%s)\" vs \"%s(%s,%s)\".",dd->name, smask->l_cptNames[0],smask->l_cptNames[1],dd->name,"NULL","NULL");
      dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
      return dmFALSE;
    }
  
    if ( !cs_uceq(smask->l_cptNames[ii], dd->cpts[ii]->name )) {//must  "X" == "X";
      sprintf( dm_errbuf, "Mask validation failure, mismatch descriptor component name of mask and source: \"%s(%s,%s)\" vs \"%s(%s,%s)\".",dd->name, smask->l_cptNames[0],smask->l_cptNames[1],dd->name,dd->cpts[0]->name,dd->cpts[1]->name);
      dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
      return dmFALSE;
    }
  }
  if ( dmDescriptorGetNoCoords( dd ) > 0 ){
    dd = dmDescriptorGetCoord( dd );
    for(ii=0; ii<smask->ndim; ii++) {
      if ( !dd->cpts[ii] && smask->w_cptNames[ii]) {
	sprintf( dm_errbuf, "Mask validation failure, mismatch descriptor component name of mask and source: \"%s(%s,%s)\" vs \"%s(%s,%s)\".",dd->name, smask->w_cptNames[0],smask->w_cptNames[1],dd->name,"NULL","NULL");
	dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
	return dmFALSE;
      }

      if ( !cs_uceq(smask->w_cptNames[ii], dd->cpts[ii]->name )) {//must  "X" == "X";
	sprintf( dm_errbuf, "Mask validation failure, mismatch descriptor component name of mask and source: \"%s(%s,%s)\" vs \"%s(%s,%s)\".",dd->name, smask->w_cptNames[0],smask->w_cptNames[1],dd->name,dd->cpts[0]->name,dd->cpts[1]->name);
	dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
	return dmFALSE;
      }
    }
  }
  dd=dd0;
  
  /*******************************************************************************************
   * ensure the mask-coordinate system is flaged along with src-coordinates's
   * otherwise, it should be updated accordingly .
   *******************************************************************************************/ 

  if ( (smask->cflag & DM_FLAG_LOGICAL) && (dd->flag & DM_FLAG_PHYS) )  {
    
    // convert the smask bounds at the PHYS coords
    dmi_mask_maskstruct_invert_bounds(smask, dmFALSE);
    smask->cflag  = dd->flag;
    
    // update mask-shape pos and coords  if any differs
    if ( myShape->flag_coord != RC_PHYSICAL  )  {
      myShape->xpos[0] = smask->xbounds[0]; 
      myShape->ypos[0] = smask->ybounds[0];
      myShape->xpos[1] = smask->xbounds[1]; 
      myShape->ypos[1] = smask->ybounds[1]; 
      myShape->flag_coord = RC_PHYSICAL;
    }
    
  } else if ( (smask->cflag & DM_FLAG_PHYS)  && (dd->flag & DM_FLAG_LOGICAL )   ) {

    // convert the smask bounds at the LOGI coords
    dmi_mask_maskstruct_invert_bounds(smask, dmTRUE);   
    smask->cflag  = dd->flag;
 
    // update mask-shape pos and coords  if any differs
    if ( myShape->flag_coord != RC_LOGICAL )  {
      myShape->xpos[0] = smask->xbounds[0]; 
      myShape->ypos[0] = smask->ybounds[0];
      myShape->xpos[1] = smask->xbounds[1]; 
      myShape->ypos[1] = smask->ybounds[1];
      myShape->flag_coord = RC_LOGICAL;
   }
 
  }

  return dmTRUE;

}

/****************************************************************** */
/*                                                                  */
/* Function:    dmi_mask_inside                                     */
/*                                                                  */
/* Description:                                                     */
/*   Bridge function for regUSER shape.                             */
/*   Determine if the input position is inside Mask shape "bounds"  */
/*   and is in an 'on' area.                                        */
/*                                                                  */
/*   Positions are assumed to be in coordinate system of Mask.      */
/*                                                                  */
/*  Input parameters:                                               */
/*    regShape shape     Mask region shape                          */
/*    double   xpos      X-axis position coordinate                 */
/*    double   ypos      Y-axis position coordinate                 */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
/*  Returns:                                                        */
/*    int result         dmTRUE(1) = 'inside' ; dmFALSE(0) = not    */
/*                                                                  */
/****************************************************************** */
int dmi_mask_inside(regShape* shape, double xpos, double ypos)
{
  int retval=0;
  double inpos[2], dpix[2];
  long dim, pixno[2];  /* byte datatype in image coords */
  long *axlens;
  int stat1, stat2;
  const double mk_round=0.5;
  dmMask *smask;
  unsigned char vbyte;
  dmBool invert;

  smask   = (dmMask*)dmi_mask_get_spec(shape);
  dim     = smask->ndim;
  axlens  = smask->axlens;

  stat1=stat2=dmSUCCESS;
  inpos[0]=xpos;
  inpos[1]=ypos;

  if( shape->flag_coord == RC_PHYSICAL ){ 
    invert=dmTRUE;
    stat1=dmi_transform_apply( smask->log2phys, &inpos[0], &dpix[0], dim, invert);   /* physical to logical */
    pixno[0]=(long)(dpix[0]+mk_round);
    pixno[1]=(long)(dpix[1]+mk_round);  

  } else  if( shape->flag_coord == RC_LOGICAL ) {
    pixno[0]=(long)( inpos[0] - shape->xpos[0]+1  ); // need "+mk_round" ?
    pixno[1]=(long)( inpos[1] - shape->ypos[0]+1  ); // need "+mk_round" ?

  } else { /* may not happen */
    double phypos[2];
    invert=dmTRUE;
    stat1=dmi_transform_apply( smask->phys2wcs, &inpos[0], &phypos[0], dim, invert); /* world to physical   */ 
    stat2=dmi_transform_apply( smask->log2phys, &phypos[0], &dpix[0], dim, invert);  /* physical to logical */ 
    pixno[0]=(long)(dpix[0] + mk_round);
    pixno[1]=(long)(dpix[1] + mk_round);  
   }

  if(stat1==dmFAILURE || stat2==dmFAILURE ){
    if(stat1==dmFAILURE)
      sprintf(dm_errbuf, "Mask masking failure, cannot convert Physical coord to Logical.");
    else
      sprintf(dm_errbuf, "Mask masking failure, cannot convert World coord to Logical.");
    dmu_err_set( DM_ERR_COORD, dm_errbuf );
    return 0;
  }

  /****************************************************************** */
  /* Simple bounds check pixel axis indexing starts at 1              */
  /****************************************************************** */
  if ( (pixno[0] < 1) || (pixno[0] > axlens[0]) || 
       (pixno[1] < 1) || (pixno[1] > axlens[1]) )
  {
    retval = 0;
  }
  else
  {
    /********************************************************************************** */
    /* get image pixel data at the 2-d pixel-no                                         */
    /* mapping 2-d pos on 1-d mask image data array                                     */
    /* (pixno[0], pixno[1]) vs smask->data                                              */
    /*                                                                                  */
    /* dmu/dmuarray.c:                                                                  */
    /*  long dmu_pixelno_from_pixel( long* Pixel, long* axes, long naxes )              */
    /*      2D -> 1D                                                                    */
    /*  void dmu_pixel_from_pixelno( long PixelNo, long* axes, long naxes, long* Pixel) */
    /*      1D -> 2D                                                                    */
    /********************************************************************************** */
    long ipix = dmu_pixelno_from_pixel( pixno, axlens, dim);
    vbyte = smask->data[ ipix -1 ];
         
    /* return pixel byte, 1|0 */
    retval = (vbyte==1 ? 1 : 0) ;

  }

  if ( shape->include == regInclude )
    return( retval );
  else 
    return( 1 - retval ); 
}

/****************************************************************** */
/*                                                                  */
/* Function:   dmi_mask_remove_blkname                              */
/*                                                                  */
/* Description:                                                     */
/*   Remove the block-name from the string contaning MASK(blkname)  */
/*                                                                  */
/*   Note: MASK(blkname) is formed by dmi_mask_tostring()           */
/*                                                                  */
/*  Input parameters:                                               */
/*    char** ss_string   Subspace string expression                 */
/*                       which is case (capital) sensitive          */
/*                                                                  */
/*  Output parameters:                                              */
/*    updated ss_string from "MASK(blkname)' to 'MASK'              */
/*                                                                  */
/*  Returns:                                                        */
/*    int:  #occurence (>=0) of updates or failure (-1)             */
/*                                                                  */
/****************************************************************** */
int dmi_mask_remove_blkname(char **ss_string)
{
  char *ss_str = *ss_string;
  char *masktag_e=")";          /* ending */
  char *maskname, *masktag_s;   /* 'MASK', 'MASK(' for start */
  
  char *ptr_l, *ptr_r;
  long noccur=0;
  long str_len, mname_len;

  if( !ss_str )
    return -1;

  
  maskname  = dmMaskName();
  mname_len = strlen(maskname);
  if( strlen(ss_str) < mname_len ){
    free(maskname);
    return 0;
  }
  
  masktag_s = (char*)malloc( (mname_len+2)*sizeof(char));
  strcat( strcpy(masktag_s,maskname), "(" );

  /* MCD NOTE: would a walking scan would be more efficient? */
  /*   Here we scan the full string each iteration.          */
  while( (ptr_l=strstr(ss_str, masktag_s) )  != NULL){
    /* pass the mask-name string but the left parenthesis, '(' */
    ptr_l += mname_len;

    /* find pointer to the ending of the tag, excluding the right parenthesis, ')'*/
    ptr_r  = strstr(ptr_l, masktag_e) + 1;

    if( !ptr_r)
      return -1; /* syntax error */
    
    /* copy the overlap the src-str to dest-str by omitting the string between '(' and ')' and including the parenthesises */
    memmove(ss_str+strlen(ss_str)-strlen(ptr_l),  ptr_r, strlen(ptr_r) );

    /* marking the end of the updated string */
    str_len=strlen(ss_str) - ( strlen(ptr_l)-strlen(ptr_r) );
    ss_str[str_len]='\0';
    
    /* record the occurance of the removings */
    noccur++;
  }

  free(masktag_s);
  free(maskname);

  /* for syntax of 
   * MASK&Circle(4120,4120,10)&MASK  or  MASK&MASK&Circle(4120,4120,10) 
   * always finalize to  MASK&Circle(4120,4120,10)
   */
  ptr_r  = strstr(ss_str,  "MASK");
  ptr_l  = strstr(ptr_r,   "&MASK");
  if( ptr_l ){
    long sz=strlen(ss_str);
    char *tmpstr=(char*)calloc(sz,sizeof(char) );
    /* get the first part, including the first MASK' */
    strncpy(tmpstr, ss_str, sz-strlen(ptr_l));
    /* get the second part without '&MASK' and cat them */
    ptr_l += 5;
    strcat(tmpstr, ptr_l);
    strcpy(ss_str,tmpstr);
    free(tmpstr);
  }
 
  return noccur;
}

/************************************************************* */
/*                                                             */
/* Function:   dmi_mask_region_from_string                     */
/*                                                             */
/* Description:                                                */
/*   Generates a Region object from region syntax string       */
/*   which includes Mask shape.  The region library does       */
/*   not recognize user shapes, so must be handled separately  */
/*                                                             */
/* Input parameters:                                           */
/*   dmDescriptor *ssd           - subspace descriptor         */
/*   const char   *dsref         - MASK reference block name   */
/*   const char   *dsval         - Region string               */
/*                                                             */
/* Output parameters:                                          */
/*   regRegion   *region         - region struct               */
/*                                                             */
/* Returns:                                                    */
/*   dmErrCode   error_code                                    */
/*                                                             */
/************************************************************* */
regRegion* dmi_mask_region_from_string( dmDescriptor* ssd, const char* dsref, const char* dsval )
{
  dmDataset* ds = NULL;
  dmBlock*   block = NULL;

  const char *pstart;              /* start of clause     */
  const char *pstop;               /* end of clause       */
  char *tmpstr;                    /* copy of clause      */
  char *ptr;                       /* ptr into copy       */
  long  size;                      /* size of clause      */
  dmBool done = UT_FALSE;          /* Loop control flag   */
  char *maskname = dmMaskName();   /* Mask shape name     */

  regRegion* region = NULL;        /* Output region       */
  regShape *shape = NULL;          /* shape ptr.          */
  regRegion *tmp_region = NULL;    /* temp. region struct */
  regMath  op;                     /* glue operation      */

  int no;                          /* block number      */
  long component;                  /* region cpt number */


  /* Get current dataset */
  block = dmi_descriptor_get_block( ssd );
  ds    = dmi_block_get_ds( block );

  /* Find referenced subspace block in dataset */
  no = dmi_block_search( ds, dsref );

  if ( no <= 0 )
  {
    sprintf( dm_errbuf, "Mask consolidate failure, referenced subspace, '%s',  is none.", dsref );
    dmu_err_set( DM_ERR_SUBSPACE, dm_errbuf );
    return region;
  }


  /* Open the block.. this is expected to be a MASK image block */
  block = dmp_block_open_mode( ds, NULL, no, dmIMAGE, NULL, "R" );

  if ( !block || (dmu_err_get() != dmSUCCESS ) )
  {
    sprintf( dm_errbuf, "Mask consolidate failure, referenced mask block, '%s', is NULL.", dsref );
    dmu_err_set( DM_ERR_SUBSPACE, dm_errbuf );
    return region;
  }

  /* Assign coord flag for mask shape */
  dmDescriptor* dd = NULL;
  if ( ssd->descType == dmSUBSPACE ) /* should be */
  {
    /* Get column corresponding to subspace */
    dd = dmi_block_search_identifier( block, ssd->name );
    if ( !dd )
      return NULL;
    block = dmi_descriptor_get_block(dd);
  }

  /* Load MASK block content to regRegion */
  region = dmi_mask_read_mask( (void*)block );
  if ( region == NULL )
  {
    sprintf( dm_errbuf, "Mask consolidate failure, cannot create read data from mask-block." );
    dmu_err_set( DM_ERR_SUBSPACE, dm_errbuf );
    return region;
  }

  /* ---------------------------------------------------------------------- */
  /* Process dsval string.. using Region parser when possible               */
  /* Expectations:                                                          */
  /*   - string is simple list of Region shapes connected by '&','|' tokens */
  /*   - '|' tokens separate clauses                                        */
  /*   - "MASK" shape, if present is first in the clause (Mask requirement) */
  /*                                                                        */
  /* ---------------------------------------------------------------------- */

  /* Create resulting region structure */
  /*   - we will add shapes to this    */

  pstart = dsval;
  while ( !done )
  {
     op = regOR;

     pstop = strchr( pstart, '|' );

     if ( !pstop || *pstop == '\0' )
     {
       done = UT_TRUE;
       size = strlen( pstart );
     }
     else
     {
       size = pstop - pstart;
     }

     /* extract this clause */
     tmpstr = cs_dup_siz( pstart, size );
     ptr = tmpstr;


     /* If clause has MASK node (at start), add it to region */
     if ( (cs_uceq_siz( ptr, maskname, strlen(maskname)) ) ||
          (cs_uceq_siz( ptr+1, maskname, strlen(maskname)) ) )
     {
       /* Mask node found.. put copy of Mask shape to region */

       shape = region->shape; //regCopyShape( mask );

       /* check exclude syntax */
       if (*ptr == '!') 
       {
         shape->include = regExclude;
         ptr++;
       }

       // regAddShape(region, op, shape);
       op = regAND;

       /* skip over mask node */
       ptr = ptr + strlen(maskname);
       if ( strlen(ptr) > 0 ){ ptr++;} /* skip operator, if any */

     }
     if ( *ptr != '\0' )
     {
       /* There is more to the clause, this should be regular Region syntax */
       /* Parse it using region parser and add results to our region.       */

       tmp_region = regParse( ptr );

       if ( !tmp_region )
       {
         sprintf( dm_errbuf, "Mask consolidate failure, region cannot parse string, '%s'.", ptr );
         dmu_err_set( DM_ERR_SUBSPACE, dm_errbuf );
         return NULL;
       }
       else
       {
         /* Transfer shapes.. do not need anything as fancy as */
         /* the regIntersect/regUnion functions.               */

         /* cut shapes from temporary region struct */
         shape = tmp_region->shape;
         tmp_region->shape = NULL;
         regFree( tmp_region );

         /* Add the chain to our region */
         regAddShape(region, op, shape );
         op = regAND;

         /* update component numbers of trailing shapes. */
         component = shape->component;
         while ( shape != NULL )
         {
            shape->component = component;
            shape = shape->next;
         }
       }
     } /* end process Region portion of clause */

     free(tmpstr);
     if ( !done ){ pstart = ++pstop; }
  }

  //dmi_mask_free(mask);
  free(maskname);

  return region;
}

/* **************************************************************************** */
/* dmi_mask_consolidate:                                                        */
/*                                                                              */
/*   Consolidates the input masks into a single MASK image.                     */
/*                                                                              */
/*   Input:                                                                     */
/*     region - regRegion containing 1 or more MASK shapes, possibly in         */
/*                combination with other region shapes and ranges               */
/*                                                                              */
/*     field  - RECTANGLE/SQUARE box defining desired MASK image bounds         */
/*                in coordinate system of filter/region                         */
/*              field[0:1] -> xmin:xmax                                         */
/*              field[2:3] -> ymin:ymax                                         */
/*                                                                              */
/*   Return:                                                                    */
/*     dmMASK struct pointer                                                    */
/*                                                                              */
/* **************************************************************************** */
dmMask* dmi_mask_consolidate( regRegion* region, double *field )
{
  int ii, jj;
  long nvals;
  long nmasks;
  double xpixbounds[2];
  double ypixbounds[2];

  regShape*  shapes = NULL;    // chain of unique MASK shapes from input region 
  regRegion* valregion = NULL; // region expression for consolidated mask values 
  dmMask*    outmask = NULL;   // resulting mask image structure.
  dmMask*    srcMask;
  const double mk_round = 0.5;
  int cflag;

  // dmverbose=5;  /* did not pass the value here ??? */

  if(dmverbose>=5)  {
    printf("\n");
    printf("#1 Get unique MASK shapes from Region. \n");
  }
  nmasks=0;
  shapes = dmi_mask_find_masks( region, &nmasks ); /* up to 2 shapes */
  if ( shapes == NULL )
  {
    if(dmverbose>=5)  
      printf("  No Masks found\n");
    return NULL;
  }

  if(dmverbose>=5)  
    printf("   - have %ld mask shapes\n", nmasks);

  if( dmverbose>=5)  {
    printf("\n");
    printf("#2 Generate region expression for determining output MASK values.\n");
  }
  valregion = dmi_mask_list_to_region( shapes );
  cflag     = ((dmMask*)valregion->shape->spec)->cflag;
  
  /* Create output Mask */
  srcMask = dmi_mask_get_spec(valregion->shape) ;
  outmask = dmi_mask_maskstruct_copy( srcMask );
  if( !outmask )
    return NULL;

  if ( field == NULL ) {
    /* use bounds of combined mask region */
    outmask->xbounds[0] = (double)valregion->xregbounds[0];
    outmask->xbounds[1] = (double)valregion->xregbounds[1];
    outmask->ybounds[0] = (double)valregion->yregbounds[0];
    outmask->ybounds[1] = (double)valregion->yregbounds[1];
  }  else {
    /* use bounds of provided field */
    outmask->xbounds[0] = field[0];
    outmask->xbounds[1] = field[1];
    outmask->ybounds[0] = field[2];
    outmask->ybounds[1] = field[3];
  }
 
  if(dmverbose>=5)  {
    printf("    - Consolidated MASK image bounds. \n");
    for ( ii = 0; ii < 2; ii++ )
      printf("      xbounds[%d] = %10.3lf  ybounds[%d] = %10.3lf \n", 
	     ii, outmask->xbounds[ii], ii, outmask->ybounds[ii] );

    printf("#3b Extract coordinate transforms.\n");
  }

  /* Determine consolidated Mask image shape in Pixel coordinates */
  /* a.k.a. resulting Mask image shape.                           */
  if( dmverbose>=5)  {
    printf("#3c Determine MASK image size. \n");
    printf("    - convert bounds to LOGICAL coords. \n");  
  }

  if ( cflag & DM_FLAG_PHYS ) 
    dmi_mask_maskstruct_invert_bounds(outmask, dmTRUE);
    

  /* transformed to or already in LOGICAL system. */
  xpixbounds[0] = outmask->xbounds[0];
  xpixbounds[1] = outmask->xbounds[1];
  ypixbounds[0] = outmask->ybounds[0];
  ypixbounds[1] = outmask->ybounds[1];

  if( dmverbose>=5)  {
    for ( ii = 0; ii < outmask->ndim; ii++ )
      printf("      xpos[%1d] = %10.3f  ypos[%1d] = %10.3f (PIXL)\n", ii, xpixbounds[ii], ii, ypixbounds[ii] );
    
    printf("    - set Mask image axis lengths. \n");
  }  

  /* check the bounds */
  if( (xpixbounds[1] >= DBL_MAX && xpixbounds[0] <= -DBL_MAX ) ||
      (ypixbounds[1] >= DBL_MAX && ypixbounds[0] <= -DBL_MAX ) ){
    sprintf(dm_errbuf, "Mask consolidate failure, out of field bounds.");
    dmu_err_set( DM_ERR_RANGE, dm_errbuf );
    regFree(valregion);
    regFreeShape(NULL, shapes);
    return  NULL;
  }

  // redefine the image size
  outmask->axlens[0] = (long)(xpixbounds[1] - xpixbounds[0] + mk_round);
  outmask->axlens[1] = (long)(ypixbounds[1] - ypixbounds[0] + mk_round); 
  
  if( dmverbose>=5)  {
    for ( ii = 0; ii < outmask->ndim; ii++ )
      printf("      Length of Axis %2d = %3ld \n", ii, outmask->axlens[ii]);
    printf("    - offset Transform if necessary. \n");  
  }
  /* - The above conversion may result in negative pixel coords...  */
  /*   shift logical2phys transform and FIELD for new image size.   */

  double offset;
  if ( xpixbounds[0] < 0 )
  {
    offset = outmask->log2phys->pix[0] - xpixbounds[0];
    if( dmverbose>=5)  
      printf("    + x-axis requires offset = %10.3f\n",offset);
    if ( offset > FLT_EPSILON )
    {
      /* Shift the Transform */
      outmask->log2phys->pix[0] += offset;
    }
  }
  if ( ypixbounds[0] < 0 )
  {
    offset = outmask->log2phys->pix[1] - ypixbounds[0];
    if( dmverbose>=5)  
      printf("    + y-axis requires offset = %10.3f\n",offset);
    if ( offset > FLT_EPSILON )
    {
      /* Shift the Transform */
      outmask->log2phys->pix[1] += offset;
    }
  }

  /* HERE WE HAVE output Mask structure with:                                  */
  /*  1. axlens     == Resulting MASK image size in pixel coords               */
  /*  2. x/ybounds  == Resulting MASK image FIELD in filter coordinates.       */
  /*  3. log2phys == Transform from Logical->Physical for consolidated MASK    */
  /*  4. phys2wcs == Transform from Physical->WCS for consolidated MASK        */
  /*  5. filter_coord = Flag for which coordinate system filter is in.         */
  /*                                                                           */
  /*  This is all the information needed to create the consolidted MASK Block  */
  /*                                                                           */
  if( dmverbose>=5)  {
    printf("\n");
    printf("#4 Create resulting image data array (byte type)\n");
    printf("   - alloc array\n");
    printf("   - set all values to 0 (OFF)\n");
  }
  nvals = 1;
  for ( ii = 0; ii < outmask->ndim; ii++ )
  {
    nvals *= outmask->axlens[ii];
  }
  outmask->data = (unsigned char*)calloc(nvals, sizeof(unsigned char));
  for(ii=0;ii<nvals;ii++)
    outmask->data[ii] = 0;

  if( dmverbose>=5)  {
    printf("\n");
    printf("#5 Populate resulting MASK image data array\n");
    printf("   - Brute force method for now..\n");
  }
  double inpos[2];
  double outpos[2];
  long ndx=0;
  
  for ( ii = 1; ii <= outmask->axlens[1]; ii++ )
  {
    inpos[1] = (double)ii;
    for ( jj = 1; jj <= outmask->axlens[0]; jj++ )
    {
      /* at Logical coord in pixels */
      inpos[0] = (double)jj;

      /* convert logi to phys coords */
      // dmi_coord_transform_image_bounds( &inpos[0], outmask->log2phys, 1, 1); the equivalent
      if( cflag & DM_FLAG_PHYS )
	dmi_transform_apply( outmask->log2phys, inpos, outpos, outmask->ndim, dmFALSE );
      else {
	outpos[0]=inpos[0];   
	outpos[1]=inpos[1];
      }
      outmask->data[ndx] = (unsigned char)regInsideRegion( valregion, outpos[0], outpos[1] );
      // 'include' flag is applied in regInsideRegion/dmi_mask_inside, 
      // but it here is appplied again in order to remove the "exclude" effects of run-time filtering
      // as the output MASK should be kept as is.
      if( !valregion->shape->include )
	outmask->data[ndx] = ( 1- outmask->data[ndx] );
      
      ndx++;
    }
  }

  if( dmverbose>=5) 
    dmi_mask_print_image(outmask->axlens[0], outmask->axlens[1], outmask->data);

  regFree(valregion);
  regFreeShape(NULL, shapes);

  return outmask;
}

/* **************************************************************************** */
/* dmi_mask_get_image_field:                                                    */
/*                                                                              */
/*   Get the DM data-descriptor fields (the most outer bounds) of an image      */
/*   in rectangular box                                                         */
/*                                                                              */
/*   Input:                                                                     */
/*        dd  - datamodel data-descriptor pointer                               */
/*                                                                              */
/*   Return:                                                                    */
/*     field  - the RECTANGLE/SQUARE box defining desired the data-descriptor   */
/*              in coordinate system of the given descriptor                    */
/*              field[0:1] -> xmin:xmax                                         */
/*              field[2:3] -> ymin:ymax                                         */
/*                                                                              */
/* **************************************************************************** */
double* dmi_mask_get_image_field(dmDescriptor* dd)
{
  int status;
  double *field=NULL;

  /* Get FIELD of image from the SKY, for example, column descriptor. */
  if( dmBlockGetType( dd->block ) != dmIMAGE )
    return NULL;

  /* ********************************************************* */
  /* this function is working on both                          */
  /* dd->descType == dmSUBSPACE                                */
  /* dd->descType == dmCOORD                                   */
  /* ********************************************************* */
  field = (double*)calloc(4,sizeof(double));
  status = dmi_descriptor_get_cpt_range( dd, 1, dmDOUBLE, (void*)&field[0],(void*)&field[1]);
  if(status != dmFAILURE)
    status = dmi_descriptor_get_cpt_range( dd, 2, dmDOUBLE, (void*)&field[2],(void*)&field[3]);

  if(status == dmFAILURE){
    free(field);
    return NULL;
  }

  if(dmverbose>=5){
    printf("\ndescriptor-name = %s",dd->name);
    printf( "\nfield(x,y): (%10.3f : %10.3f)\t(%10.3f : %10.3f)\n", field[0], field[1], field[2], field[3] );
  }

  return field;

}
/* **************************************************************************** */
/* dmi_mask_maskstruct_free:                                                    */
/*                                                                              */
/*  free all array memories allocated but the dmBlock pointer                   */
/*  which is to be freed elseplace                                              */
/*                                                                              */
/*   Input:                                                                     */
/*     smasks - MASK struct pointer                                             */
/*   Output:  - none                                                                    */
/*                                                                              */
/* **************************************************************************** */
void dmi_mask_maskstruct_free(dmMask *smask)
{
  long ii, ndim;

  if(smask) {
    if (smask->axlens )
      free(smask->axlens);
    if (smask->data ) 
      free(smask->data);
    if(smask->xbounds)
      free(smask->xbounds);
    if(smask->ybounds)
      free(smask->ybounds);
    if(smask->log2phys)
      dmi_transform_free(smask->log2phys);
   if(smask->phys2wcs)
      dmi_transform_free(smask->phys2wcs);

   ndim=smask->ndim;
   if(smask->l_cptNames){
     for(ii=0;ii<ndim;ii++)
       free(smask->l_cptNames[ii]);
     free(smask->l_cptNames);
   }
   if(smask->w_cptNames){
     for(ii=0;ii<ndim;ii++)
       free(smask->w_cptNames[ii]);
     free(smask->w_cptNames);
   }

   if(smask->blkname)
     free(smask->blkname);

    free(smask);
  }
}

/* **************************************************************************** */
/* dmi_mask_find_masks:                                                         */
/*                                                                              */
/*   Scans input region for regMASK shapes.                                     */
/*   Creates a regSHAPE chain containing copies of the unique set of            */
/*   regMASK shapes found.                                                      */
/*                                                                              */
/*   Input:                                                                     */
/*     region - regRegion to scan                                               */
/*     nmasks - only check if the region has MASK if nmasks==dmMAXLEN           */
/*                                                                              */
/*   Output:                                                                    */
/*     nmasks - How many masks in chain                                         */
/*                                                                              */
/*   Return:                                                                    */
/*     result - chain of regSHAPE objects                                       */
/*              from input region                                               */
/*                                                                              */
/* **************************************************************************** */
regShape* dmi_mask_find_masks( regRegion* region, long* nmasks )
{

  regShape* iptr = NULL;      // iterator for input shapes
  regShape* optr = NULL;      // iterator for output shapes
  regShape* result = NULL;    // return variable
  regShape* shape  = NULL;    // copy 

  dmBool found;
  long count = 0;

  iptr = region->shape; 
  /* just check if containing one MASK at least */ 
  if( *nmasks == dmMAXLEN ){  
    *nmasks=0; /* initiated to 0 */
    while ( iptr != NULL )
      { 
	if ( iptr->type == regMASK ){
	  *nmasks=1;
	  return NULL;
	} else 
	  iptr = iptr->next;
      }
  }

  /* otherwise check how many MASKs are contained and return the allocated shapes */ 
  while ( iptr != NULL )
    { 
    if ( iptr->type == regMASK )
    {
      // MASK shape.. check if we already have this one.
      found = dmFALSE;
      optr = result;
      while ( optr && !found )
      {
        if ( dmi_mask_maskstruct_isequal(optr->spec, iptr->spec ) )
	{
          found = dmTRUE;
          break;
        }
        optr = optr->next;
      }
      if ( !found )
      {
        // New MASK shape.. copy and add to set
        shape = dmi_mask_copy( iptr );
	if ( result == NULL )
          result = shape;
        else
        {
          optr = result;
          while ( optr->next != NULL ){ optr = optr->next; }
          optr->next = shape;
        }
        count++;
      }
    }
    iptr = iptr->next;
  }

  /* set output variable with number of masks found */
  *nmasks = count;
  return( result );
}

/* **************************************************************************** */
/* dmi_mask_list_to_region:                                                     */
/*                                                                              */
/*   Generate a regRegion object containing the region expression needed to     */
/*   determine the values for a combined list of Masks.                         */
/*   NOTE: This is NOT an intersection, the resulting field covers the full     */
/*         bounds of all provided masks.  However, for overlapping areas, it    */
/*         is the intersection of values which will result.                     */
/*                                                                              */
/*   Input:                                                                     */
/*     inmasks - chain of regShape objects                                      */
/*                                                                              */
/*   Return:                                                                    */
/*     result - chain of regShape objects                                       */
/*              from input region                                               */
/*                                                                              */
/* **************************************************************************** */
regRegion* dmi_mask_list_to_region( regShape* inmasks )
{
  int ii,jj;
  long ndx, nmasks;
  regShape* iptr = NULL;
  regShape* tmp  = NULL;
  regRegion* result = NULL;

  regShape** maskarr = NULL;  // array of mask shapes
  regShape** fieldarr = NULL; // array of field shapes

  double xbounds[2];
  double ybounds[2];

  /* Count number of masks */
  nmasks = 0;
  iptr = inmasks;
  while ( iptr )
  { 
    nmasks++;
    iptr = iptr->next;
  }

  if ( nmasks == 0 )
    return NULL;


  /* Generate expression for determining output MASK values. */
  if( dmverbose>=5)
    printf("   - Define FIELD for each mask.\n");


  /* Make array of masks and associated fields.. easier to work with here */
  maskarr  = calloc( nmasks, sizeof( regShape* ) );
  fieldarr = calloc( nmasks, sizeof( regShape* ) );

  ndx = 0;
  iptr = inmasks;
  while ( iptr )
  {
    maskarr[ndx] = iptr;
    xbounds[0]=xbounds[1]=0.0;
    ybounds[0]=ybounds[1]=0.0;
    /* generate regRECTANGLE shape for FIELD of each mask */
    if ( dmi_mask_extent( iptr, &xbounds[0], &ybounds[0] ) != dmSUCCESS )
      return NULL;
 
    tmp = regCreateShape( NULL, regAND, regRECTANGLE, regInclude, xbounds, ybounds, 2, NULL, NULL, 0, 0 );
    if ( tmp ){
      if( dmverbose>=5) {
	printf("     FIELD\n");
	printf("       nPoints = %ld\n", tmp->nPoints );
	for ( ii = 0; ii < tmp->nPoints; ii++ )
	  printf("       xpos[%d] = %10.3lf  ypos[%d] = %10.3lf \n", ii, tmp->xpos[ii], ii, tmp->ypos[ii] );
      }
    } else {
      return NULL;
    }

    fieldarr[ndx] = tmp;

    ndx++;
    iptr = iptr->next;
  }
  if( dmverbose>=5)
    printf("   - create empty region. \n");

  /* Create empty region and assign the bounds */
  result = regCreateEmptyRegion();

  /* For each mask, generate a clause of MASK&!FIELD(<other masks>) */
  /* serving values for non-overlapping areas.                      */
  if( dmverbose>=5)
    printf("   - Load with clauses for non-overlapping areas. \n");
  for ( ii = 0; ii < nmasks; ii++ )
  {
      
    tmp = regCopyShape( maskarr[ii] );
    regAddShape( result, regOR, tmp );
    for ( jj = 0; jj < nmasks; jj++ )
    {
      if ( ii != jj )
      {
 	tmp = regCopyShape( fieldarr[jj] );
	tmp->include = regExclude;
	regAddShape( result, regAND, tmp );
      }
    }
  }

  /* For each mask (i), generate a clause of MASK(i)&MASK(j)) j>i */
  /* Serving values for overlapping areas                        */
  if( dmverbose>=5)
    printf("   - Load with clauses for overlapping areas. \n");
  for ( ii = 0 ; ii < nmasks; ii++ )
  {
     for ( jj = ii+1; jj < nmasks; jj++ )
     {
       tmp = regCopyShape( maskarr[ii] );
       regAddShape( result, regOR, tmp );
       tmp = regCopyShape( maskarr[jj] );
       regAddShape( result, regAND, tmp );
     }
  }

  if( dmverbose>=5) {
    char* tmpstr = regAllocComposeRegion( result ) ;
    printf("   REGION: %s \n", tmpstr );
    free(tmpstr);
  }

  /* Clean up */
  if ( maskarr )
  {
    free(maskarr);
    maskarr = NULL;
  }
  if ( fieldarr )
  {
    for ( ii = 0; ii < nmasks; ii++ )
    { 
      if ( fieldarr[ii] )
        regFreeShape( NULL, fieldarr[ii] );
    }
    free(fieldarr);
    fieldarr = NULL;
  }

  return( result ) ;
}


/* **************************************************************************** */
/* dmi_mask_get_transforms:                                                     */
/*                                                                              */
/*   Extract Transform information from DM block.                               */
/*                                                                              */
/*   Input:                                                                     */
/*     block  - DM block.                                                       */
/*                                                                              */
/*   Output:                                                                    */
/*     log2phys = Logical to Physical axis Transform from the given block       */
/*     phys2wcs = Physical to World axis Transform from the given block         */
/*                                                                              */
/*   Return:                                                                    */
/*     <none>                                                                   */
/*                                                                              */
/* **************************************************************************** */
void dmi_mask_get_transforms( dmDescriptor *dd0, dmCoordTransform** log2phys, dmCoordTransform** phys2wcs )
{
  dmCoordTransform* t;        // coord Transform 
  dmDescriptor *dd;

  dd=dd0;
  if( dd->flag & DM_FLAG_LOGICAL ) 
    dd = dmDescriptorGetCoord( dd );

  t = (dmCoordTransform*)dmi_descriptor_get_aux( dd );
  if( t  ) { 
    *log2phys = dmi_coord_transform_copy( NULL, t, 2L );
    
    if( (*log2phys)->name ) 
      free( (*log2phys)->name );
    (*log2phys)->name = cs_dup(dd->name);  // name == SKY
    
    if( dmverbose>=5)
      printf("      + L2P Transform == %s\n", dmi_coord_get_transform_type(dd) );
  }

  if ( dmDescriptorGetNoCoords( dd ) > 0 )
  {
    /* extract next coordinate level (World)     */
    dd = dmDescriptorGetCoord( dd );
    t = (dmCoordTransform*)dmi_descriptor_get_aux( dd );
    if( !t  ) 
      return;
    *phys2wcs = dmi_coord_transform_copy( NULL, t, 2L );
  
    if( (*phys2wcs)->name ) 
      free( (*phys2wcs)->name );
    (*phys2wcs)->name = cs_dup(dd->name); // name == EQPOS
    if(t->radesys){
      if( (*phys2wcs)->radesys ) 
	free( (*phys2wcs)->radesys );
      (*phys2wcs)->radesys = cs_dup(t->radesys);
    }
  }

  if( dmverbose>=5)
    printf("      + P2W Transform == %s\n", dmi_coord_get_transform_type(dd0) );
  
   return;
}
/****************************************************************************** */
/* dmi_mask_copy_coordsys:                                                           */
/*                                                                              */
/*  Set/transform the data-descriptor coords                                    */ 
/*                                                                              */
/*   Input:                                                                     */
/*        dd  - datamodel data-descriptor pointer                               */
/*   newmask  - MASK struct pointer                                             */
/*                                                                              */
/*                                                                              */
/* Returns:                                                                     */
/*   dmErrCode   error_code                                                     */
/*                                                                              */
/****************************************************************************** */
dmErrCode  dmi_mask_copy_coordsys(dmDescriptor* dd, dmMask *newmask)
{
 
 /*
 * HHE NOTE:
 *
 * Following steps have been worked:                                            
 * 
 * First Create Linear-transform coords
 *  dd_coord=dmArrayCreateAxisGroup(dd, vector-name, data-type, data-unit, vector-cpts-names, dim)
 *  dmCoordSetTransform_d(dd_coord, pix,val,delt)
 * Then add non-Linear-transform coords
 *  dd_coord=dmCoordCreate_d(dd_coord, ra-dec-name, deg,radec-cpts,...)
 *  dmCoordSetTransform_d( dd_coord, ....)
 */

  dmDescriptor* dd_coord;
  dmErrCode  status;
  long  dim;
  dmCoordTransform *coord;
  char **cptNames;

  if(!dd || !newmask)  
    return dmFAILURE;
  

  status=dmSUCCESS;
  dim = newmask->ndim;
  coord  = newmask->log2phys; 
  cptNames=newmask->l_cptNames;
  dd_coord = dmArrayCreateAxisGroup(dd, coord->name, dmDOUBLE, "", cptNames, dim);
  if(dd_coord==NULL)
    return dmFAILURE;     
  status = dmCoordSetTransform_d( dd_coord,  (double*)coord->pix, (double*)coord->val, (double*)coord->delt, dim);
  if(status != dmSUCCESS)
    return status;

  coord    = newmask->phys2wcs;
  if(coord){
    cptNames = newmask->w_cptNames;
    dd_coord=dmCoordCreate_d(dd_coord, coord->name, "deg", cptNames, dim, coord->type,
			     (double*)coord->pix, (double*)coord->val, (double*)coord->delt,NULL);     
    if(dd_coord==NULL)
      return dmFAILURE;
    else     
      return dmCoordSetTransform_d( dd_coord,  (double*)coord->pix, (double*)coord->val, (double*)coord->delt,dim);
  
  }
  
  return status;

}


/* **************************************************************************** */
/* dmi_mask_copy_maskstruct:                                                    */
/*                                                                              */
/*   Allocate memory and copy mask-struct information from the src mstruct.     */
/*                                                                              */
/*   Input:                                                                     */
/*     dmMask*  - source mask-struct pointer                                    */
/*                                                                              */
/*   Output:                                                                    */
/*     <none>                                                                   */
/*                                                                              */
/*   Return:                                                                    */
/*     dmMadk* -  copy of the source's                                          */
/*                                                                              */
/* **************************************************************************** */
dmMask *dmi_mask_maskstruct_copy(dmMask *src)
{
  dmMask* out;
  int ii, npix;
  
  out = dmi_mask_maskstruct_alloc(src->ndim);
  out->log2phys = NULL;
  out->phys2wcs = NULL;

  if( src->log2phys ) {
    out->log2phys = dmi_transform_copy( src->log2phys, src->ndim);
    out->log2phys->name = cs_dup(src->log2phys->name);
  }

  if( src->phys2wcs ){
    out->phys2wcs = dmi_transform_copy( src->phys2wcs, src->ndim);
    if(src->phys2wcs->name)
      out->phys2wcs->name = cs_dup(src->phys2wcs->name);
  }

  for(ii=0; ii<src->ndim; ii++){
    out->axlens[ii]=src->axlens[ii];

    out->xbounds[ii]=src->xbounds[ii];
    out->ybounds[ii]=src->ybounds[ii];

    if(src->l_cptNames[ii])
      out->l_cptNames[ii] = cs_dup(src->l_cptNames[ii]);
    if(src->w_cptNames[ii])
      out->w_cptNames[ii] = cs_dup(src->w_cptNames[ii]);
  }

  npix = out->axlens[0] * out->axlens[1];
  out->data = (unsigned char*)calloc(npix, sizeof(unsigned char));
  for(ii=0; ii<npix;ii++)
    out->data[ii] = src->data[ii];
 
  if(src->blkname)
    out->blkname = cs_dup(src->blkname);

  out->cflag = src->cflag;
  
  return out;

}


/* **************************************************************************** */
/* dmi_mask_allocate_maskstruct:                                                */
/*                                                                              */
/*   Allocate memory to a mask-struct pointer                                   */
/*                                                                              */
/*   Input:                                                                     */
/*     int ndim  - number of dimentions                                         */
/*                                                                              */
/*   Output:                                                                    */
/*     <none>                                                                   */
/*                                                                              */
/*   Return:                                                                    */
/*     dmMadk* -  memory allocated mask-struct                                  */
/*                                                                              */
/* **************************************************************************** */
dmMask *dmi_mask_maskstruct_alloc(int ndim)
{
  dmMask*    outmask; 

  /* Create output Mask */
  outmask = (dmMask*)calloc( 1, sizeof( dmMask ) );

  outmask->ndim    = ndim;
  outmask->axlens  = ( long   *)calloc( outmask->ndim, sizeof(long)   );
  outmask->xbounds = ( double *)calloc( outmask->ndim, sizeof(double) );
  outmask->ybounds = ( double *)calloc( outmask->ndim, sizeof(double) );

  outmask->l_cptNames = ( char **)calloc( outmask->ndim, sizeof(char*) );
  outmask->w_cptNames = ( char **)calloc( outmask->ndim, sizeof(char*) );

  outmask->blkname  = NULL;
  outmask->log2phys = NULL;
  outmask->phys2wcs = NULL;
  outmask->data     = NULL;
  outmask->cflag    = -1;

  return outmask;

}

/* **************************************************************************** */
/* dmi_mask_convert:                                                            */
/*                                                                              */
/*   Extract info from DM block and pass them a mask-struct                     */
/*                                                                              */
/*   Input:                                                                     */
/*     dmBlock* blk  - DM block pointer                                         */
/*                                                                              */
/*   Output:                                                                    */
/*     <none>                                                                   */
/*                                                                              */
/*   Return:                                                                    */
/*     dmMask* -  memory allocated and info filled mask-struct                  */
/*                                                                              */
/* **************************************************************************** */
dmMask* dmi_mask_convert( dmBlock *blk)
{
  dmDescriptor *dd, *dd0;
  dmDataType dtype; // associated with dmDescriptor
  double *field=NULL;
  dmMask *smask;
  long npix, ndim, ngrp, ii;
  long *urpix;
  void *values;
  void *nulval;


  dd    =  dmImageGetDataDescriptor(blk);  
  ndim  = dmi_descriptor_get_array_dimensions( dd, &urpix ); /* bounds on logical coordinates */
  if( ndim != 2 ) // must be 2
    return NULL;
  
  // assign values to the mask-struct memebers
  smask = dmi_mask_maskstruct_alloc(ndim);
  smask->axlens[0]=urpix[0];
  smask->axlens[1]=urpix[1];
 
  ngrp = dmi_descriptor_get_no_groups(dd);
  if( ngrp==1 ) {
    dd0  = dmArrayGetAxisGroup( dd, 1 );     // initiated at PHYSICAL coords
    if( !dd0 ) {
      // on the logical coords
      // logical start from 0.5
      smask->xbounds[0] = 0.5;
      smask->xbounds[1] = smask->axlens[0]+0.5;
      smask->ybounds[0] = 0.5;
      smask->ybounds[1] = smask->axlens[1]+0.5;
      dd0=dd;
    } else {
      // get bounds on physical coordinates 
      field = dmi_mask_get_image_field(dd0);
      if( field == NULL) 
	return NULL;
      smask->xbounds[0] = field[0];
      smask->xbounds[1] = field[1];
      smask->ybounds[0] = field[2];
      smask->ybounds[1] = field[3];
      free(field);
    }
    dmi_mask_get_transforms( dd0, &smask->log2phys, &smask->phys2wcs);
    smask->cflag = dd0->flag;

    if (smask->log2phys){
      if(smask->cflag & DM_FLAG_LOGICAL )
	dd0 = dmDescriptorGetCoord( dd0 );
      for(ii=0;ii<dd0->dim;ii++)
	smask->l_cptNames[ii] = cs_dup(dd0->cptNames[ii]); //cs_dup("X") or "Y";
    }

    if(smask->phys2wcs){
      dd0 = dmDescriptorGetCoord( dd0 );
      for(ii=0;ii<dd0->dim;ii++)
	smask->w_cptNames[ii] = cs_dup(dd0->cptNames[ii]); //cs_dup("RA") or "DEC"; 
    }   
  } else if( ngrp==2 ){ //deal with (pi,pha) image type
    dmCoordTransform* tl;
    dmCoordTransform* tw;
    for(ii=1; ii<=ngrp; ii++){
      dd0 = dmi_descriptor_get_group_no(dd,ii);
      if (dd0){
	// get bounds on physical coordinates 
	field = dmi_mask_get_image_field(dd0);
	if( field == NULL) 
	  return NULL;
	if( ii==1 ){
	  smask->xbounds[0] = field[0];
	  smask->xbounds[1] = field[1];
	} else {
	  smask->ybounds[0] = field[0];
	  smask->ybounds[1] = field[1];
	}
	free(field);
      } else {
	// on the logical coords
	smask->xbounds[0] = 0.5;
	smask->xbounds[1] = smask->axlens[0]+0.5;
	smask->ybounds[0] = 0.5;
	smask->ybounds[1] = smask->axlens[1]+0.5;
	dd0=dd;
      }
 
      if(ii==1) {
	dmi_mask_get_transforms( dd0, &smask->log2phys, &smask->phys2wcs);
	smask->cflag = dd0->flag;
      } else { //ii==2 
	if( smask->cflag != dd0->flag ){
	  printf("ERROR: group-1 and -2 coords are different.\n");
	  return NULL;
	}
	tl=NULL;
	tw=NULL;
	dmi_mask_get_transforms( dd0, &tl, &tw);
	smask->log2phys->val[ii-1]  = tl->val[0];
  	smask->log2phys->pix[ii-1]  = tl->pix[0];
 	smask->log2phys->delt[ii-1] = tl->delt[0];

	// form vector name, say "(pi,pha)",
	char tname[dmMAXSTR];
	strcpy(tname, "(");
	strcat(tname,smask->log2phys->name);
	strcat(tname,",");
	strcat(tname,tl->name);
	strcat(tname, ")");
	free(smask->log2phys->name);
	smask->log2phys->name= cs_dup(tname);
      }

      if (smask->log2phys){
	if(dd0->flag & DM_FLAG_LOGICAL )
	  dd0 = dmDescriptorGetCoord( dd0 );
	if(dd0->dim==1) // must be 1
	  smask->l_cptNames[ (ii-1)*(ngrp-1)] = cs_dup(dd0->cptNames[0]); //cs_dup("pi") or "pha";
      }

      if(smask->phys2wcs){
	dd0 = dmDescriptorGetCoord( dd0 );
	if(dd0->dim==1)
	  smask->w_cptNames[ (ii-1)*(ngrp-1)] = cs_dup(dd0->cptNames[0]); //cs_dup("RA") or "DEC"; 
      }
    }
    if(tl)  dmi_transform_free(tl);
    if(tw)  dmi_transform_free(tw);
  }

  smask->blkname = cs_dup( dmi_block_get_name(blk) );
  // note that dtype=DT_UNIT1  for *dd
  //           dtype=DT_REAL*  for *dd0 
  values = dmi_mask_get_image_array(blk, &dtype, &npix); // dtype==byte

  if(values == NULL)
    return NULL;

  nulval = dmi_mask_get_null( dd ); 
  smask->data = (unsigned char*)dmi_mask_convert_values(values, dtype, npix, nulval);
  if(nulval)
    free( nulval );
  
  return smask;
	
}

/* **************************************************************************** */
/* dmi_mask_print_maskstruct:                                                   */
/*                                                                              */
/*   Print mask-struct memeber's value                                          */
/*                                                                              */
/*   Input:                                                                     */
/*     dmMask*  - mask-struct content pointer                                   */
/*                                                                              */
/*   Output:                                                                    */
/*     std output on screen                                                     */
/*                                                                              */
/*   Return:                                                                    */
/*     <none>                                                                   */
/*                                                                              */
/* **************************************************************************** */
void dmi_mask_maskstruct_print(dmMask *smask)
{

  if( smask->log2phys ){
    printf("\nLogical2Phys Transform in type %s",  smask->log2phys->type);
    printf("\n%10s(%8s,%8s):", smask->log2phys->name,smask->l_cptNames[0],smask->l_cptNames[1]);
    dmi_mask_print_transform( smask->log2phys );
  }
  if( smask->phys2wcs ){
    printf("\nPhys2World Transform in type %s",  smask->phys2wcs->type);
    printf("\n%10s(%8s,%8s):", smask->phys2wcs->name, smask->w_cptNames[0],smask->w_cptNames[1]);
    dmi_mask_print_transform( smask->phys2wcs );
  }

  printf("\n --- Mask Block, '%s', Image Data --- ", smask->blkname);
  printf("\nAxies length   (Xpixels, Ypixels)   = (%10ld, %10ld)",  smask->axlens[0], smask->axlens[1]);
  printf("\nAxies bounds (xmin:xmax, ymin:ymax) = (%10.3f:%10.3f, %10.3f:%10.3f)",  smask->xbounds[0],smask->xbounds[1],smask->ybounds[0],smask->ybounds[1]);
  dmi_mask_print_image(smask->axlens[0], smask->axlens[1], smask->data);

  return;

}

/* **************************************************************************** */
/* dmi_mask_print_transform:                                                    */
/*                                                                              */
/*                                                                              */
/*   Input:                                                                     */
/*     t - transformation struct                                                */
/*                                                                              */
/* **************************************************************************** */
void dmi_mask_print_transform( dmCoordTransform* t)
{
  int ii;

  if(t==NULL)
    return;
  for(ii=0; ii<2; ii++){
    if(ii==0)
      printf("\n%20s: (val, delt, pix)[%d] = (%10.3f, %13.7f, %10.3f)",  
	     t->name, ii, t->val[ii], t->delt[ii], t->pix[ii]);
    else
      printf("\n%20s  (val, delt, pix)[%d] = (%10.3f, %13.7f, %10.3f)",  
	     "", ii, t->val[ii], t->delt[ii], t->pix[ii]);
  }
}

/****************************************************************************** */
/* dmi_mask_print_image:                                                        */
/*                                                                              */
/*   print image to the size upto 99x99 pixels                                  */ 
/*                                                                              */
/*   Input:                                                                     */
/*     xlen - the lenth along X-axis                                            */
/*     ylen - the lenth along Y-axis                                            */
/*     data - the data array                                                    */
/*                                                                              */
/* **************************************************************************** */
void dmi_mask_print_image(long xlen, long ylen, unsigned char *data)
{
  long ndx, jj, ii;
  if( xlen >=99 || ylen >=99 ){
    printf("\n Skip printing image, limitted to the size of 99x99 .\n");
    return;
  }
  ndx=0;

  for ( jj = ylen-1; jj >= -2; jj-- )
  {
    if ( jj >= 0 )
      printf("%3d |", (short)jj+1 );
    else
      printf("     " );

    for ( ii = 0; ii < xlen; ii++ )
    {
      if ( jj >= 0 )
      {
	ndx = jj*xlen + ii ; 
        printf("%3d", (short)data[ndx] );
      }
      else if ( jj == -1 )
        printf("---");
      else if ( jj == -2 )
        printf("%3d", (short)ii+1 );
    }
    printf("\n");
  }
}


/*************************************************
  union  multi-masks 
 ************************************************/

int dmi_mask_union_masks(long ntotal, regShape **allmasks, regRegion **allregions);

int dmi_mask_union_updatemasks(regShape *newmask, long ntotal, regRegion **allregions);
int dmi_mask_union_bounds(regShape *newmask, long ntotal, regShape **allmasks);
int dmi_mask_union_data(regShape *newmask, long ntotal, regRegion **allregions);
int dmi_mask_union_transform(regShape *newmask, long ntotal, regShape **allmasks);
int dmi_mask_union_extent(double *cxpos, double *cypos, double *sxpos, double *sypos);
dmBool dmi_mask_union_isequal(long ntotal, regShape **allmasks);

int dmi_mask_union_masks(long ntotal, regShape **allmasks, regRegion **allregions )
{
  // NOTE:
  // by Helen 7/15/2015
  // this function does not handle  dmmerge run with multi-masks, exist + run-time
  // but seems it takes care of dmcopy run with the multi components

  regShape *newmask;
  
  // take care of dmcopy with multi components
  if( dmi_mask_union_isequal( ntotal, allmasks) ) 
    return 1;

  // create a new mask from the copy
  newmask = regCopyShape(allmasks[0]);

  // update  new mask for min/max bounds from original masks
  dmi_mask_union_bounds( newmask, ntotal, allmasks );
  
  // update  new mask for transform from original masks
  if( !dmi_mask_union_transform( newmask, ntotal, allmasks) ) { 
    regFreeShape(NULL, newmask);
    return 0;
  }

  // update the newmask data
  if( !dmi_mask_union_data(newmask, ntotal, allregions ) ) {
    regFreeShape(NULL, newmask);
    return 0;
  }

  if( !dmi_mask_union_updatemasks(newmask, ntotal, allregions) ) {
    regFreeShape(NULL, newmask);
    return 0;
  }

  regFreeShape(NULL, newmask);
  return 1;
}

// take care of dmcopy with multi components
dmBool dmi_mask_union_isequal(long ntotal, regShape **allmasks)
{
  long ii;
  regShape *iptr1, *iptr2;

  for( ii=1; ii<ntotal; ii++ ){
    iptr1 = allmasks[0];
    iptr2 = allmasks[ii];
    while( iptr1 && iptr2) {
      if( !dmi_mask_isequal( iptr1, iptr2 ) )
	return dmFALSE;
      iptr1=iptr1->next;
      iptr2=iptr2->next;
      if( (iptr1 && !iptr2) || (!iptr1 && iptr2) )
	return dmFALSE;
    }
  }
  return  dmTRUE;
}

// here does not handle  dmmerge multi-masks, exist + run-time
int dmi_mask_union_bounds(regShape *newmask, long ntotal, regShape **allmasks)
{
  long ii, nvals;
  dmMask* smask;
  
  // update  new mask for min/max bounds from original masks
  for( ii=0; ii<ntotal; ii++ ) 
    dmi_mask_union_extent( &newmask->xpos[0], &newmask->ypos[0], allmasks[ii]->xpos, allmasks[ii]->ypos);
  

  smask = dmi_mask_get_spec(newmask);
  // update new mask bounds
  for(ii=0; ii<smask->ndim; ii++){
    smask->xbounds[ii]= newmask->xpos[ii];
    smask->ybounds[ii]= newmask->ypos[ii];
  }

  smask->axlens[0] = newmask->xpos[1] - newmask->xpos[0];
  smask->axlens[1] = newmask->ypos[1] - newmask->ypos[0];

  nvals=  smask->axlens[0] * smask->axlens[1]; 
  if( smask->data )
    free(smask->data);
  smask->data = (unsigned char*)calloc(nvals, sizeof(unsigned char) );
  return 1;
}

int dmi_mask_union_updatemasks(regShape *newmask, long ntotal, regRegion **allregions) 			
{
  double fx[2] = { DT_REAL8_MIN, DT_REAL8_MAX };
  double fy[2] = { DT_REAL8_MIN, DT_REAL8_MAX };

  regShape* orimask;
  regShape* orishapes;
  regRegion* oriregion;
  long ii;


  // update  with unioned newmask for original masks
  for( ii=0; ii<ntotal; ii++ ) {
    oriregion = allregions[ii];

    /* get the mask address from the original region */
    orimask  = regGetShapeNo(oriregion, 1);

    // make sure 'mask' is in the first position of the shape-lister in the origina region
    if ( orimask->type != regMASK )  
      return 0;
 
    // get the rest shape pointer or orimask->next pointer or more next list
    orishapes = orimask->next;
    while ( orishapes && orishapes->type==regMASK ) {
      orishapes = orishapes->next;
    }

    // set ori-shape point to a new mask with all newmask's content
    oriregion->shape =  regCopyShape(newmask);
    oriregion->shape->region = oriregion;

    // add the ori-shapes after the 'mask' shape if any  
    if ( orishapes ) {          
      if(orimask->component != orishapes->component)
	regAddShape(oriregion, regOR, orishapes);
      else
	regAddShape(oriregion, regAND, orishapes);
    }

    // free the old-mask shape
    regFreeShape(NULL, orimask);

    // set new bounds from the new combo
    regExtent(oriregion, fx,fy, oriregion->xregbounds, oriregion->yregbounds);
  }

  return 1;
}

int dmi_mask_union_data(regShape *newmask, long ntotal, regRegion **allregions)
{
  // update the contents
  long ndx, nvals, result, ii;
  dmMask* smask;

  smask = dmi_mask_get_spec(newmask);
  nvals=  smask->axlens[0] * smask->axlens[1]; 
  if( !smask->data )
    smask->data = (unsigned char*)calloc(nvals, sizeof(unsigned char));

  // initialized to zeor
  for(ndx=0; ndx<nvals; ndx++)
    smask->data[ndx] = 0.0;
  long pos[2];
  double inpos[2];
  double outpos[2];

  for(ndx=1; ndx<=nvals; ndx++) {
    dmu_pixel_from_pixelno( ndx, smask->axlens, smask->ndim, &pos[0] );

    /* convert logi to phys coords */
    if( smask->cflag & DM_FLAG_PHYS ) {
      inpos[0]=pos[0];
      inpos[1]=pos[1];
      dmi_transform_apply( smask->log2phys, inpos, outpos, smask->ndim, dmFALSE );
    } else {
      outpos[0]=pos[0];   
      outpos[1]=pos[1];
    }

    // look into  all regions
    for( ii=0; ii<ntotal; ii++){
      result = (unsigned char)regInsideRegion( allregions[ii], outpos[0], outpos[1] );
      if( result ) {
	smask->data[ndx] = 1;
	break; // to next ndx loop
      }
    }
    
  }

  return 1;
}

int dmi_mask_union_transform(regShape *newmask, long ntotal,  regShape **allmasks)
{
  dmMask *smask;  
  dmCoordTransform* log2phys_1=NULL;	
  dmCoordTransform* phys2wcs_1=NULL;
  dmCoordTransform* log2phys_2=NULL;
  dmCoordTransform* phys2wcs_2=NULL;
  long ii, jj;
  double diff_dlt, diff_val, diff_pix;
  float torl=1.e-5;   //FLT_EPSILON    TBdecided
 
  // newmask is input and also output
  smask = (dmMask*)dmi_mask_get_spec( newmask);
  log2phys_1 = smask->log2phys;
  phys2wcs_1 = smask->phys2wcs;

  for( jj=0; jj<ntotal; jj++ ) {
    
    /* get coords transformations */
    smask = (dmMask*)dmi_mask_get_spec( allmasks[jj]);
    log2phys_2 = smask->log2phys;
    phys2wcs_2 = smask->phys2wcs;


   /*******************************************************************************************
    validate the same coords transformation params between Mask and event/filter descriptors 
    
    y = offset + scale * ( x - reference )
    val[2]: offset
    delt[2]: scale     
    pix[2]: reference
   *******************************************************************************************/
    for( ii = 0; ii < smask->ndim; ii++ ) {
      /* for Linear tyep: make sure the scales and references are the same   */   
      diff_pix = fabs(  log2phys_1->pix[ii] - log2phys_2->pix[ii] );
      diff_dlt = fabs(  log2phys_1->delt[ii] - log2phys_2->delt[ii]); 
    
      if( diff_pix >= FLT_EPSILON || diff_dlt >= FLT_EPSILON  ){ 
	if( diff_dlt >= FLT_EPSILON )   /* validate b==b' */                
	  sprintf( dm_errbuf, "Mask coordinate failure, mismatch '%s' transformation scales for '%s'.",log2phys_1->type, log2phys_1->name);
	else   	              /* validate A-A'==integer */
	  sprintf( dm_errbuf, "Mask coordinate failure, mismatch '%s' transformation references for '%s'.",log2phys_1->type, log2phys_1->name);
	
	dmu_err_set( DM_ERR_COORD, dm_errbuf );
	return  0;
      }
    

      if( phys2wcs_1 &&  phys2wcs_2 ) {
	/* for TAN tyep: ignore checking the reference  */
	diff_val = fabs(phys2wcs_1->val[ii]  - phys2wcs_2->val[ii]);
	diff_dlt = fabs(phys2wcs_1->delt[ii] - phys2wcs_2->delt[ii]);
	diff_pix = fabs(phys2wcs_1->pix[ii]  - phys2wcs_2->pix[ii]);
	if( diff_val >= torl || diff_dlt >= FLT_EPSILON  ) {
	  if( diff_val >= torl )
	    sprintf( dm_errbuf, "Mask coordinate failure, mismatch '%s' transformation offsets for '%s'.", phys2wcs_1->type, phys2wcs_1->name);
	  else
	    sprintf( dm_errbuf, "Mask coordinate failure, mismatch '%s' transformation deltas for '%s'.", phys2wcs_1->type, phys2wcs_1->name);
	  
	  dmu_err_set( DM_ERR_COORD, dm_errbuf );
	  return  0;
	}
      } // end of  if( phys2wcs_1 &&  phys2wcs_2 )
    
    } // end of for( ii = 0; ii < dim; ii++ ) 

    for( ii = 0; ii < smask->ndim; ii++ ) {
      /* update the newmask shape's offset value with the smallest value of the shape2 */
      if( log2phys_1->val[ii]  > log2phys_2->val[ii] )
	log2phys_1->val[ii]  = log2phys_2->val[ii];
    }
  } // end of for( jj=0; 

  return 1;
}
/* Increase cx to include sx */
int dmi_mask_union_extent(double *cxpos, double *cypos, double *sxpos, double *sypos)
{
/* Here we are doing the union, so the ranges get bigger */
    int retval = 0;
    if (sxpos[0] < cxpos[0]) {
	cxpos[0] = sxpos[0];
	retval = 1;
    }
    if (sxpos[1] > cxpos[1]) {
	cxpos[1] = sxpos[1];
	retval = 1;
    }
    if (sypos[0] < cypos[0]) {
	cypos[0] = sypos[0];
	retval = 1;
    }
    if (sypos[1] > cypos[1]) {
	cypos[1] = sypos[1];
	retval = 1;
    }
    if (retval) {
	if (cxpos[0] > cxpos[1])
	    cxpos[0] = cxpos[1];
	if (cypos[0] > cypos[1])
	    cypos[0] = cypos[1];
    }
    return retval;
}

/***********************************************************************
 *
 * is filter binning matching mask's coords scale?
 * the matching rule: binning scale factor is the multiple mask's scale.
 ***********************************************************************/
dmBool dmi_mask_match_bin(dmSubspace* subspace, dmBool isbinned)
{
  dmFilter     *filter     = NULL;  /* Value group of source component  */
  dmDescriptor *ss_dd      = NULL;
  dmDescriptor *col_dd     = NULL;
  long colOffset, CptNo;
  long n, nmasks, ntotal;
  regShape *masks;
  regShape **allmasks;
  regRegion **allregions;
  long stat, ii;

  n = subspace->ncpts;
  if( n<=0 )
    return dmTRUE;
  else if ( n>1 ) {

    // note:!!
    // all GTIs or all subspace compnoents have mask filters
    // we need to deal one only, TBD.
    // MCD NOTE 20151118: I have commented out this message..
    //   This method appears to be run whether there are masks or not,
    //   but only has action when there are masks.  The message is 
    //   printed in either case, which is confusing when there are 
    //   no masks involved.
    //   The message caused a couple DM test failures and shows 
    //   on the screen of many ciao tests.
    //printf("\nWARNING: number of subspace components (%ld) encountered (in '%s'). Will continue Only for the 1st component.\n", n, __FILE__);
    n=1;
  }
  
  allmasks   = calloc( n, sizeof( regShape* )  );
  allregions = calloc( n, sizeof( regRegion* ) );
  for(CptNo=1; CptNo <= n; CptNo++) {
    allmasks[CptNo-1]   = NULL;
    allregions[CptNo-1] = NULL;
  }

  // iterate columns
  stat=dmTRUE;
  for(colOffset=0; colOffset< subspace->ncols; colOffset++){
    ss_dd = subspace->cols[colOffset];
    ntotal=0;
    // iterate components of the column
    for(CptNo=1; CptNo <= n; CptNo++) {
       filter = dmi_subspace_col_get_filter( ss_dd, CptNo);
      if ( !filter || !filter->region )
	continue;

      nmasks=0;
      // create new regShape*
      masks = dmi_mask_find_masks( filter->region, &nmasks) ;

      if(nmasks == 0) 
	continue;
      allmasks[ntotal]   = masks;     // memory should be fred before exit
      allregions[ntotal] = filter->region ;
      ntotal++;
    }  // end of  for(CptNo=1;

    if( ntotal == 0) 
      continue;
    
    //
    // take account of the bin-factor which is just parsed but may not be updated yet.
    // search a descriptor list for a named descriptor.  Search   
    // of the the descriptors' themselves, their vector components,
    // their coords, and their groups.  
    //

    // get the dmDescriptor from source dmBlock given the component-name string
    dmBlock *src_blk = ss_dd->block ;
    col_dd = dmi_descriptor_list_search( (const dmDescriptor**)src_blk->cols, src_blk->ncols, ss_dd->name );
    // try other way
    if( !col_dd ) { // take cares of (pi,pha), or the arbitary component combos 
      col_dd = dmi_block_search_raw_col_identifier( src_blk, ss_dd->name ); 
      if(!col_dd) {// take cares of (pi,pha), or the arbitary component combos 
	sprintf( dm_errbuf, "Error: Can't find data descriptor (%s) for rebinning mask image.", ss_dd->name);
	dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
	return dmFALSE;
      }
    }

    stat=dmTRUE;
    long multis[2];
    for( CptNo=0; CptNo < ntotal; CptNo++) {
      // add shape->next checkup TBD
      // which run-time or existing one?
      regShape* shape =  allmasks[CptNo];
      while (shape) {
	//	stat = dmi_mask_match_coords( shape, col_dd, col_dd->name);
	//	stat = dmi_mask_match_coords( shape, col_dd, ss_dd->name);
	// okay if exact match, otherwise check binsacle/mscale multiple below
	//	if( !stat && isbinned ) {
	// check multiple 
	stat = dmi_mask_match_scale_multiple(shape, col_dd, &multis[0]);

	if( !isbinned && stat ) {    // having no binning, need eactly match, okay multiple=1
	  if( multis[0] != 1 || multis[1] != 1) 
	    stat = dmFALSE; // error out
	} 
	
	if( !stat )
	  break;

	shape=shape->next;
      }
      if( !stat )
	break;
    }

    //set to ZERO status for union next column's
    for(ii=0; ii < ntotal; ii++) {
      if( allmasks[ii] )
	regFreeShape(NULL, allmasks[ii]);
      allmasks[ii]   = NULL;
    }
    
    if(!stat)
      break;
  }
  
  free(allmasks);
  free(allregions);

  return stat;

}

/***********************************************************************
 *
 * is filter binning scale matching mask's coords scale?
 * or is binning scale factor the multiple mask's scale?
 ***********************************************************************/
dmBool dmi_mask_match_scale_multiple(regShape *myShape, dmDescriptor* dd, long *multiples)
{
  int ii;
  dmMask *smask;
  double multi;

  dmCoordTransform* log2phys_1=NULL;
  dmCoordTransform* log2phys_2=NULL;
  dmCoordTransform* phys2wcs_2=NULL;

  smask = (dmMask*)dmi_mask_get_spec(myShape);

  /* get coords transformations */
  log2phys_1 = smask->log2phys;
  dmi_mask_get_transforms( dd, &log2phys_2, &phys2wcs_2);

  /* for Linear tyep, ignore checking the world  */   
  for( ii = 0; ii < smask->ndim; ii++ ) {
    if ( log2phys_1->delt[ii] <  FLT_EPSILON )
      log2phys_1->delt[ii] = 1.0;
    multi = log2phys_2->delt[ii] / log2phys_1->delt[ii]; 

    if ( fabs( multi - floor(multi) ) > FLT_EPSILON ) {
      sprintf( dm_errbuf, "Error: Bin step (%g) is not a multiple of Mask's step (%g).", log2phys_2->delt[ii], log2phys_1->delt[ii] );
      dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
      return dmFALSE;
    }
    multiples[ii]=(long)multi;

  }

  return dmTRUE;
}

/***********************************************************************
 *
 * bin mask data in binning spec,
 *    iteratly rebin  all mask associated with in subspace
 *    update mask struct contents accordingly.
 ***********************************************************************/
int dmi_mask_rebin_mask(dmSubspace* subspace, dmBinSpec*  binspec)
{
  dmFilter     *filter     = NULL;  /* Value group of source component  */
  dmDescriptor *ss_dd      = NULL;
  long colOffset, CptNo;
  long n, nmasks, stat;
  regShape *shape;
     
  if( !binspec->axis || !binspec->axis->vals || binspec->axis->vals[0][2]==1)
    return dmSUCCESS;

  n = subspace->ncpts;
  if( n<=0 )
    return dmSUCCESS;

  // iterate columns
  stat=dmSUCCESS;
  for(colOffset=0; colOffset< subspace->ncols; colOffset++){
    ss_dd = subspace->cols[colOffset];

    // iterate components of the column
    for(CptNo=1; CptNo <= n; CptNo++) {
      filter = dmi_subspace_col_get_filter( ss_dd, CptNo);
      if ( !filter || !filter->region )
	continue;

      nmasks=0;
      // create new regShape*
      shape = dmi_mask_find_masks( filter->region, &nmasks) ;

      if(nmasks == 0) 
	continue;

      if( shape )
	regFreeShape(NULL, shape);

      // rebing mask data  /* otherwise check how many MASKs are contained and return the allocated shapes */ 
      shape =  filter->region->shape;
      while ( shape && shape->type==regMASK ){ 
	stat = dmi_mask_rebin_data( shape, binspec);
	if ( stat != dmSUCCESS)
	  return stat;
	shape=shape->next;
      }
    } // end of  for(CptNo=1;    
  }

  return stat;
}

/***********************************************************************
 *
 * bin mask data in binning spec, using binspec's bounds and an 
 *     effective binning scale combined binsepc's with mask's scale.
 * update mask struct contents accordingly.
 *
 * reference: 
 *  ./filter/long dmBinningParse()
 ***********************************************************************/
int dmi_mask_rebin_data(regShape *shape, dmBinSpec*  binspec)
{
  long stat;
  dmMask *smask;
  long n1,n2;             // original image extent
  unsigned char *data;    // original or unbinned data
  double scalemx, scalemy;
  long npix;              // = n1*n2;

  unsigned char *ndata;        // regridded or rebinned data
  long nn1, nn2;               // binned image extent
  double scalebinx, scalebiny; // bin factor or scale
  double minbinx, maxbinx,minbiny,maxbiny; // binspec (minbinx:maxbinx:scalebinx, minbiny:maxbiny:scalebiny)
  long nnpix;              // = nn1*nn2;
 
  double scalex, scaley;   // bin factor after combined
  long xoffset0, yoffset0; // diffs between two bounds
  long cellsz_x, cellsz_y;

  /* References
   * dmi/dmparselib.c: 
   *  L2798-2869: force the bin range to the pixel boundaries
   *  L3822 Image rebin; use existing image type 

   * misc/dmbuffer.c: L2258
   * image filtering with [opt full]
   *
   *  dmi_adjust_binning_edges( char* name, long vcpt, double* vals, int is_integer )
   *  dmi_adjust_binning_edges( "", 2, double* vals, 1 )
   *  L2488 dmi_axis_steps( double* vals, long type ) : rounding for steps, partial consider

   * DM only support linear binning,  DM_BINS_N 
#define DM_BINS_NONE -1
#define DM_BINS_N    1
#define DM_BINS_LOG  2
#define DM_BINS_COMPACT 4
#define DM_BINS_FILE 8

   binspec->axis->vals (min:max) in physical coords

   */
  stat = dmSUCCESS;

  smask   = dmi_mask_get_spec(shape);
  scalemx = smask->log2phys->delt[0];
  scalemy = smask->log2phys->delt[1];
  // get the final scale factors from the multiples
  //  scalex==1 or scaley==1 exit without binning
  //  else binning
  //  bscale=4, mscale=2, the final scale=2
  if( scalemx < FLT_EPSILON)
    scalemx=1.0;
  if( scalemy < FLT_EPSILON)
    scalemy=1.0;

  scalebinx = binspec->axis->vals[0][2];
  scalebiny = binspec->axis->vals[1][2];

  scalex = scalebinx/scalemx;
  scaley = scalebiny/scalemy;

  // no need to do anything further if the effective scale factor is unit
  if( fabs(scalex-1.0)< FLT_EPSILON && fabs(scaley-1.0)< FLT_EPSILON)
    return dmSUCCESS;

  // in the physical coords
  minbinx =  binspec->axis->vals[0][0];
  maxbinx =  binspec->axis->vals[0][1];
  minbiny =  binspec->axis->vals[1][0];
  maxbiny =  binspec->axis->vals[1][1];

  // BOUNDS:
  // In  ./filter/dmparselib.c:dmi_parse_bin_filter(), L2069
  //  it does not check if the bin range is within the current evtent/masks's
  // for the case of 
  //  binspec   sky=2000.0:4500:2
  //  mask bounds x=3449.5:3650.5:1, y=3649.5:3950.5:1
  //  xoffset0 and yoffset0 below will  be negative numbers
  // To deal with this the new-data points at the edges,
  //  value out of mask bounds is set to zero, rataining the
  //  new image extents with the specified bin range

  xoffset0 = (long)( minbinx - smask->log2phys->val[0] + FLT_EPSILON);
  yoffset0 = (long)( minbiny - smask->log2phys->val[1] + FLT_EPSILON);


  data = smask->data;
  n1 = smask->axlens[0];
  n2 = smask->axlens[1];
  npix = n1*n2; 


  // new number data array in binned image scope
  // igore any partial binning or ignore edges either on X- or Y-axis
  double vals[3];
  vals[0]=minbinx;
  vals[1]=maxbinx;
  vals[2]=scalex;
  nn1 = dmi_axis_steps( vals, binspec->axis->type);
  vals[0]=minbiny;
  vals[1]=maxbiny;
  vals[2]=scaley;
  nn2 = dmi_axis_steps( vals, binspec->axis->type);
  nnpix = nn1*nn2;
  ndata =  calloc( nnpix, UT_SZ_CHAR); // SIZEOF_UNSIGNED_CHAR

  // cell size along x-y axies
  cellsz_x = (long)( scalex + FLT_EPSILON);
  cellsz_y = (long)( scaley + FLT_EPSILON);

//
// loop over pixel image, or in logical coords system
// assuming full binning
//
  long ii,jj,ix,iy, indx, nindx;
  unsigned char data_pt, ndata_pt;
  long xoffset, yoffset;   
  long ii0,jj0;
  long total_counts, ntotal_counts;

  // initiate new data array to zeros
  for(ii=0; ii<nnpix;ii++)
    ndata[ii]=0;
  nindx = 0;
  total_counts =0;
  ntotal_counts=0;

  for( jj=0; jj<nn2; jj++ ){ // along y-axis pixel image
    yoffset =  jj*(cellsz_y *n1) + yoffset0;
    for(ii=0; ii<nn1; ii++ ){ // along x-axis pixel image
      xoffset = ii*cellsz_x + xoffset0;

      ndata_pt = 0.0;
      // loop the data in the binning cell bounds in  celllsz_x, cellsz_y
      for(iy=0; iy<cellsz_y; iy++) {
	for(ix=0; ix<cellsz_x; ix++) {
	  indx = ix + iy*n1 + xoffset + yoffset;
	  if (indx  < npix && indx >= 0)
	    data_pt = data[indx];
	  else
	    data_pt = 0;
	  
	  // data_pt =0||1 matters ?
	  ndata_pt += data_pt;
	} // end loop cellx
      } // end loop celly

      // definition on the binned about "on" and "off"
      // 'on' if all is '1' otherwise 'off'
      total_counts += ndata_pt;
      if (ndata_pt == cellsz_y*cellsz_x) {    // image 'on' always true
      // if (ndata_pt != 0) {  // image 'on' always true
	ndata_pt=1;     
	ntotal_counts++;
      } else 
	ndata_pt=0;

      if ( nindx < nnpix)
	ndata[nindx] = ndata_pt; 
      nindx++;
      ii0=ii;
    } // end loop ii
    jj0=jj;
  } // end loop jj


  // confirm the looping logicals
  // look up if any partial pixels to be rounded up into the last celll
  ii0 +=1;
  jj0 +=1;

  dmBool isPrinted=dmFALSE;
  if(ii0 != nn1 && dmverbose>=1 ) {  // it looks DM did not parse dmverbose down
    printf( "\n WARNING: mask partial binning at X-axis: nn1=%2d vs ii=%d.", (int)nn1, (int)ii0 );
    isPrinted=dmTRUE;
  }
  if(jj0 != nn2 && dmverbose>=1 ) {
    printf( "\n WARNING: mask partial binning at Y-axis: nn2=%2d vs jj=%d.\n", (int)nn2, (int)jj0 );
    isPrinted=dmTRUE;
  }

  double partialx= (minbinx + nn1*scalex + FLT_EPSILON ) - maxbinx;
  double partialy= (minbiny + nn2*scaley + FLT_EPSILON ) - maxbiny;
  if(  fabs(partialx ) > FLT_EPSILON   && dmverbose>=1 ) {
    printf( "\nWARNING: Mask partial binning (or more pixels) at X-axis: %9.3f.", partialx );
    isPrinted=dmTRUE;
  }
  if(  fabs( partialy ) > FLT_EPSILON  && dmverbose>=1 ){
    printf( "\nWARNING: mask partial binning (or more pixels) at Y-axis: %9.3f.", partialy );
    isPrinted=dmTRUE;
  } 
  if( isPrinted)
    printf( "\n");

  if( dmverbose>=1 ){
    printf( "\npre-bin  Mask total 'on's: %ld", total_counts);
    printf( "\npost-bin Mask total 'on's: %ld\n", ntotal_counts);
  }
 
  // update smask contents
  free (smask->data);
  smask->data = ndata;
  smask->axlens[0] =  nn1;
  smask->axlens[1] =  nn2;
  smask->log2phys->delt[0] = scalebinx; // the final scale factor should be recorded here
  smask->log2phys->delt[1] = scalebiny; // the final scale factor should be recorded here
  smask->log2phys->val[0] = minbinx; 
  smask->log2phys->val[1] = minbiny; 
  smask->xbounds[0] = minbinx;
  smask->xbounds[1] = maxbinx;
  smask->ybounds[0] = minbiny;
  smask->ybounds[1] = maxbiny;

  // update the coords reference values
  dmDescriptor* dd=binspec->axis->dd;
  dmCoordTransform* log2phys=NULL;
  dmCoordTransform* phys2wcs=NULL;
  /* get coords transformations */
  if( dd )
    dmi_mask_get_transforms( dd, &log2phys, &phys2wcs);
  if(log2phys && log2phys->scaled ){
    // update the coords reference values
    smask->log2phys->pix[0] = log2phys->pix[0];
    smask->log2phys->pix[1] = log2phys->pix[1];
  }

  return stat;
  
}
